from datetime import datetime

import neo4j

from icij_worker.constants import (
    NEO4J_TASK_MANAGER_EVENT_EVENT,
    NEO4J_TASK_MANAGER_EVENT_NODE,
    NEO4J_TASK_MANAGER_EVENT_NODE_CREATED_AT,
)
from icij_worker.event_publisher.event_publisher import EventPublisher
from icij_worker.objects import ManagerEvent
from icij_worker.task_storage.neo4j_ import Neo4jStorage


class Neo4jEventPublisher(Neo4jStorage, EventPublisher):

    async def _publish_event(self, event: ManagerEvent):
        async with self._task_session(event.task_id) as sess:
            await _publish_event(sess, event)

    @property
    def driver(self) -> neo4j.AsyncDriver:
        return self._driver


async def _publish_event(sess: neo4j.AsyncSession, event: ManagerEvent):
    as_json = event.json(exclude_none=True)
    await sess.execute_write(
        _publish_manager_event_tx, as_json, created_at=event.created_at
    )


async def _publish_manager_event_tx(
    tx: neo4j.AsyncTransaction, event_as_json: str, created_at: datetime
):
    create_manager_event = f"""
CREATE (event:{NEO4J_TASK_MANAGER_EVENT_NODE} {{
    {NEO4J_TASK_MANAGER_EVENT_EVENT}: $eventAsJson,
    {NEO4J_TASK_MANAGER_EVENT_NODE_CREATED_AT}: $createdAt 
}})
RETURN event"""
    await tx.run(create_manager_event, eventAsJson=event_as_json, createdAt=created_at)



from __future__ import annotations

import asyncio
import functools
import inspect
import logging
import os
import socket
import threading
import traceback
from abc import abstractmethod
from asyncio import TimerHandle
from contextlib import AbstractAsyncContextManager, asynccontextmanager
from copy import deepcopy
from datetime import datetime, timezone
from inspect import isawaitable
from typing import (
    Any,
    Callable,
    Dict,
    Optional,
    Tuple,
    Type,
    TypeVar,
    final,
)

from typing_extensions import Self

from icij_common.pydantic_utils import safe_copy
from icij_worker import AsyncApp, ResultEvent, Task, TaskError, TaskState
from icij_worker.app import RegisteredTask, supports_progress
from icij_worker.event_publisher.event_publisher import EventPublisher
from icij_worker.exceptions import (
    MaxRetriesExceeded,
    RecoverableError,
    TaskAlreadyCancelled,
    TaskAlreadyReserved,
    UnknownTask,
    UnregisteredTask,
    WorkerTimeoutError,
)
from icij_worker.objects import (
    CancelEvent,
    CancelledEvent,
    ErrorEvent,
    ProgressEvent,
    TaskUpdate,
    WorkerEvent,
)
from icij_worker.routing_strategy import RoutingStrategy
from icij_worker.utils import RegistrableFromConfig
from icij_worker.worker.process import HandleSignalsMixin

logger = logging.getLogger(__name__)

C = TypeVar("C", bound="WorkerConfig")
WE = TypeVar("WE", bound=WorkerEvent)


class Worker(
    RegistrableFromConfig,
    EventPublisher,
    HandleSignalsMixin,
    AbstractAsyncContextManager,
):
    def __init__(
        self,
        app: AsyncApp,
        worker_id: Optional[str] = None,
        *,
        group: Optional[str],
        handle_signals: bool = True,
        teardown_dependencies: bool = False,
    ):
        # If worker are run using a thread backend then signal handling might not be
        # required, in this case the signal handling mixing will just do nothing
        HandleSignalsMixin.__init__(self, logger, handle_signals=handle_signals)
        self._app = app
        if worker_id is None:
            worker_id = self._create_worker_id()
        self._id = worker_id
        self._group = group
        self._teardown_dependencies = teardown_dependencies
        self._graceful_shutdown = True
        self._loop = asyncio.get_event_loop()
        self._work_forever_task: Optional[asyncio.Task] = None
        self._work_once_task: Optional[asyncio.Task] = None
        self._watch_cancelled_task: Optional[asyncio.Task] = None
        self._already_exiting = False
        self._current: Optional[Task] = None
        self._worker_events: Dict[Type[WE], Dict[str, WE]] = {CancelEvent: dict()}
        self._config: Optional[C] = None
        self._cancel_lock = asyncio.Lock()
        self._current_lock = asyncio.Lock()
        self._successful_exit = False
        self._timeout_exc: Optional[WorkerTimeoutError] = None
        self._timeout_callback_handle: Optional[TimerHandle] = None

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        return self._loop

    @property
    def cancel_lock(self) -> asyncio.Lock:
        return self._cancel_lock

    @property
    def _routing_strategy(self) -> RoutingStrategy:
        return self._app.routing_strategy

    @functools.cached_property
    def id(self) -> str:
        return self._id

    @final
    def work_forever(self):
        # This is a bit cosmetic but the sync one is useful to be run inside Python
        # worker multiprocessing Pool, while async one is more convenient for testing
        # Start watching cancelled tasks
        self._work_forever_task = self._loop.create_task(self._work_forever_async())
        self._loop.run_until_complete(self._work_forever_task)

    @final
    async def _work_forever_async(self):
        async with self:
            self.info("started working...")
            try:
                await self._work_forever()
            except asyncio.CancelledError:  # Shutdown let's not reraise
                self.info("worker cancelled, shutting down...")
            except KeyboardInterrupt:  # Shutdown let's not reraise
                pass
            except Exception as e:
                self.exception("error occurred while consuming: %s", _format_error(e))
                self.info("will try to shutdown gracefully...")
                raise e
            finally:
                self.info(
                    "finally stopped working, nothing lasts forever, "
                    "i'm out of this busy life !"
                )

    @final
    async def _work_forever(self):
        while True:
            try:
                self._work_once_task = asyncio.create_task(self._work_once())
                await self._work_once_task
            except TaskAlreadyReserved:
                # This part is won't happen with AMQP since it will take care to
                # correctly forward one task to one worker
                self.info("tried to consume an already reserved task, skipping...")
                continue
            except TaskAlreadyCancelled:
                # This part is won't happen with AMQP since it will take care to
                # correctly forward one task to one worker
                self.info("tried to consume a cancelled task, skipping...")
                continue

    def _get_worker_event(self, task: Task, type: Type[WE]) -> WE:
        try:
            return self._worker_events[type][task.id]
        except KeyError as e:
            raise UnknownTask(task_id=task.id, worker_id=self._id) from e

    @final
    def logged_name(self) -> str:
        return self.id

    @property
    def graceful_shutdown(self) -> bool:
        return self._graceful_shutdown

    @final
    async def _work_once(self):
        async with self.ack_cm:
            self._current = await task_wrapper(self, self._current)

    @final
    async def consume(self) -> Task:
        task = await self._consume()
        self.debug('Task(id="%s") locked', task.id)
        async with self._current_lock:
            self._current = task
            progress = 0.0
            update = {"progress": progress, "state": TaskState.RUNNING}
            self._current = safe_copy(task, update=update)
        event = ProgressEvent.from_task(self._current)
        await self.publish_event(event)
        return self._current

    @final
    @property
    def ack_cm(self):
        if self._late_ack:
            return self._late_ack_cm()
        return self._early_ack_cm()

    @final
    @asynccontextmanager
    async def _early_ack_cm(self):
        current_id = None
        try:
            await self.consume()
            await self.acknowledge(self._current)
            yield
            current_id = self._current.id
            async with self._current_lock:
                self._clear_current()
            self.info('Task(id="%s") successful !', current_id)
        except asyncio.CancelledError as e:
            async with self.cancel_lock, self._cancel_lock:
                await self._handle_task_cancellation(e)
        except (TaskAlreadyCancelled, TaskAlreadyReserved) as e:
            # Let this bubble up and the worker continue without recording anything
            raise e
        except RecoverableError as e:
            async with self.cancel_lock, self._current_lock:
                if self._current is not None:
                    await self._handle_error(e, fatal=False)
                else:
                    self.exception(
                        'Task(id="%s") recoverable error while publishing success',
                        current_id,
                    )
        except Exception as fatal_error:  # pylint: disable=broad-exception-caught
            async with self._current_lock, self._cancel_lock:
                if self._current is not None:
                    # The error is due to the current task, other tasks might success,
                    # let's fail this task and keep working
                    await self._handle_error(fatal_error, fatal=True)
                    return
            # The error was in the worker's code, something is wrong that won't change
            # at the next task, let's make the worker crash
            raise fatal_error

    @final
    @asynccontextmanager
    async def _late_ack_cm(self):
        current_id = None
        try:
            await self.consume()
            yield
            current_id = self._current.id
            async with self._current_lock:
                await self.acknowledge(self._current)
                self._clear_current()
            self.info('Task(id="%s") successful !', current_id)
        except asyncio.CancelledError as e:
            async with self.cancel_lock, self._current_lock:
                await self._handle_task_cancellation(e)
        except (TaskAlreadyCancelled, TaskAlreadyReserved) as e:
            # Let this bubble up and the worker continue without recording anything
            raise e
        except RecoverableError as e:
            async with self._current_lock, self._cancel_lock:
                if self._current is not None:
                    await self._handle_error(e, fatal=False)
                else:
                    self.exception(
                        'Task(id="%s") recoverable error while publishing success',
                        current_id,
                    )
        except Exception as fatal_error:  # pylint: disable=broad-exception-caught
            async with self._current_lock, self._cancel_lock:
                if self._current is not None:
                    # The error is due to the current task, other tasks might success,
                    # let's fail this task and keep working
                    await self._handle_error(fatal_error, fatal=True)
                    return
            # The error was in the worker's code, something is wrong that won't change
            # at the next task, let's make the worker crash
            raise fatal_error

    async def _handle_error(self, error: BaseException, fatal: bool):
        task = self._current
        if isinstance(error, RecoverableError):
            error = error.args[0]
        exceeded_max_retries = isinstance(error, MaxRetriesExceeded)
        retries_left = self._current.retries_left - 1 if not fatal else 0
        if exceeded_max_retries:
            self.exception('Task(id="%s") exceeded max retries', task.id)
        elif fatal:
            self.exception('Task(id="%s") fatal error during execution', task.id)
        else:
            self.exception('Task(id="%s") encountered error', task.id)
        task_error = TaskError.from_exception(error)
        await self.publish_error_event(task_error, self._current, retries_left)
        if self._late_ack:
            if fatal:
                await self._negatively_acknowledge_(self._current)
            else:
                await self._acknowledge(self._current)
        self._clear_current()

    async def _handle_cancel_event(self, cancel_event: CancelEvent):
        async with self.cancel_lock, self._current_lock:
            if self._current is None:
                self.info(
                    'Task(id="%s") completed before cancellation was effective, '
                    "skipping cancellation!",
                    cancel_event.task_id,
                )
                return
            if self._current.id != cancel_event.task_id:
                self.info(
                    'worker switched to Task(id="%s") before it could cancel'
                    ' Task(id="%s"), skipping cancellation!',
                    self._current.id,
                    cancel_event.task_id,
                )
                return
            self.info(
                'received cancel event for Task(id="%s"), cancelling it !',
                cancel_event.task_id,
            )
            self._work_once_task.cancel()
        await self._work_once_task

    async def _handle_task_cancellation(self, e: asyncio.CancelledError):
        if self._timeout_exc is not None:
            self.exception(
                'Task(id="%s") consumer exceeded allocated timeout', self._current.id
            )
            fatal = not self._app.config.recover_from_worker_timeout
            await self._handle_error(self._timeout_exc, fatal=fatal)
            return
        if self._worker_cancelled:
            # we just raise and the Worker.__aexit__ will take care of
            # handling worker shutdown gracefully
            raise e
        if self._current is None:
            logger.info(
                "task cancellation was ask but task was acked or nacked"
                " in between, discarding cancel event !"
            )
            return
        event = self._get_worker_event(self._current, CancelEvent)
        update = {"state": TaskState.QUEUED if event.requeue else TaskState.CANCELLED}
        self._current = safe_copy(self._current, update=update)
        self.info('Task(id="%s") cancellation requested !', self._current.id)
        await self._publish_cancelled_event(event)
        self._clear_current()

    @property
    def _late_ack(self) -> bool:
        return self._app.config.late_ack

    @final
    async def acknowledge(self, task: Task):
        self.info('Task(id="%s") acknowledging...', task.id)
        await self._acknowledge(task)
        self.info('Task(id="%s") acknowledged', task.id)

    @abstractmethod
    async def _acknowledge(self, task: Task): ...

    @final
    async def _negatively_acknowledge_(self, task: Task):
        if not self._late_ack:
            raise ValueError("can't negatively acknowledge with early ack")
        self.info("negatively acknowledging Task(id=%s)...", task.id)
        await self._negatively_acknowledge(task)
        # This is ugly but makes tests easier
        # (shutdown doesn't try to requeue the task again)
        if task is self._current:
            self._clear_current()
        self.info("Task(id=%s) negatively acknowledged !", task.id)

    @abstractmethod
    async def _negatively_acknowledge(self, nacked: Task): ...

    @abstractmethod
    async def _consume(self) -> Task: ...

    @abstractmethod
    async def _consume_worker_events(self) -> WorkerEvent: ...

    @final
    async def publish_result_event(self, result: Any, task: Task):
        self.debug('Task(id="%s") publish result event', task.id)
        result_event = ResultEvent.from_task(task, result)
        await self.publish_event(result_event)

    @final
    async def publish_error_event(self, error: TaskError, task: Task, retries: int):
        self.debug('Task(id="%s") publish error event', task.id)
        error_event = ErrorEvent.from_task(
            task, error, retries_left=retries, created_at=datetime.now(timezone.utc)
        )
        await self.publish_event(error_event)

    @final
    async def _publish_progress(self, progress: float, task: Task):
        async with self._current_lock:
            self._current = safe_copy(task, update={"progress": progress})
        event = ProgressEvent.from_task(self._current)
        await self.publish_event(event)

    @final
    def _publish_progress_sync(self, progress: float, task: Task):
        self._loop.call_soon(self._publish_progress, progress, task)

    async def _publish_cancelled_event(self, cancel_event: CancelEvent):
        update = {
            "state": TaskState.QUEUED if cancel_event.requeue else TaskState.CANCELLED
        }
        self._current = safe_copy(self._current, update=update)
        cancelled_event = CancelledEvent.from_task(
            self._current, requeue=cancel_event.requeue
        )
        if not cancel_event.requeue and self._late_ack:
            await self.acknowledge(self._current)
        await self.publish_event(cancelled_event)

    @final
    def parse_task(self, task: Task) -> Tuple[Callable, Tuple[Type[Exception], ...]]:
        registered = _retrieve_registered_task(task, self._app)
        recoverable = registered.recover_from
        task_fn = registered.task
        if supports_progress(task_fn):
            publish_progress = functools.partial(self._publish_progress, task=task)
            task_fn = functools.partial(task_fn, progress=publish_progress)
        return task_fn, recoverable

    @final
    async def _watch_worker_events(self):
        try:
            while True:
                worker_event = await self._consume_worker_events()
                event_task_id = worker_event.task_id
                existing_events = self._worker_events[worker_event.__class__]
                already_received = event_task_id in existing_events
                not_processing = self._current is None
                if already_received or not_processing:
                    continue
                existing_events[event_task_id] = worker_event
                if isinstance(worker_event, CancelEvent):
                    await self._handle_cancel_event(worker_event)
                else:
                    raise ValueError(f"unexpected event type {worker_event}")
        except Exception as fatal_error:
            async with self._current_lock, self._cancel_lock:
                if self._current is not None:
                    await self._handle_error(fatal_error, fatal=True)
            raise fatal_error
        except asyncio.CancelledError as e:
            logger.info("cancelling cancelled task watch !")
            raise e

    @final
    def check_retries(self, task: Task, original_exc: Exception):
        self.info(
            '%sTask(id="%s"): try %s/%s',
            task.name,
            task.id,
            task.retries_left,
            task.max_retries,
        )
        if task.retries_left <= 0:
            raise MaxRetriesExceeded(
                f"{task.name}(id={task.id}): max retries exceeded > {task.max_retries}"
            ) from original_exc

    @final
    def __enter__(self):
        self._loop.run_until_complete(self.__aenter__())

    @final
    async def __aenter__(self) -> Self:
        await self._aenter__()
        # Start watching cancelled tasks
        self._watch_cancelled_task = self._loop.create_task(self._watch_worker_events())
        return self

    @final
    def bind_worker_timeout(self):
        registered = _retrieve_registered_task(self._current, self._app)
        group = registered.group
        if group is None:
            return
        group = self._app.task_group(group.name)
        if group is None:
            return
        timeout = group.timeout_s
        if timeout is None:
            return
        msg = f'Task(id="{self._current}") exceeded allocated timeout of {timeout}s'
        exc = WorkerTimeoutError(msg)
        self._timeout_callback_handle = self.loop.call_later(
            timeout, self.worker_timeout_callback, exc
        )

    @final
    def worker_timeout_callback(self, exc: WorkerTimeoutError):
        self._timeout_exc = exc
        if self._work_once_task is not None:
            self._work_once_task.cancel()

    async def _aenter__(self):
        pass

    @final
    def __exit__(self, exc_type, exc_value, tb):
        self._loop.run_until_complete(self.__aexit__(exc_type, exc_value, tb))

    @final
    async def __aexit__(self, exc_type, exc_value, tb):
        # dependencies might be closed while trying to gracefully shutdown
        if not self._already_exiting:
            self._already_exiting = True
            if (
                self._watch_cancelled_task is not None
                and not self._watch_cancelled_task.cancelled()
            ):
                logger.info("terminating cancel event loop ...")
                self._watch_cancelled_task.cancel()
                # Wait for cancellation to be effective
                try:
                    await self._watch_cancelled_task
                except asyncio.CancelledError:
                    pass
                self._watch_cancelled_task = None
                logger.info("cancel event loop terminated !")
            # Let's try to shut down gracefully
            await self.shutdown()
            # Clean worker dependencies only if needed, dependencies might be share in
            # which case we don't want to tear them down
            if self._teardown_dependencies:
                self.info("cleaning worker dependencies...")
                await self._aexit__(exc_type, exc_value, tb)
            self._successful_exit = True

    async def _aexit__(self, exc_type, exc_val, exc_tb):
        pass

    @final
    async def _shutdown_gracefully(self):
        async with self._current_lock, self._cancel_lock:
            if self._current is not None:
                cancel_event = CancelEvent.from_task(self._current, requeue=True)
                await self._publish_cancelled_event(cancel_event)
                self._clear_current()

    @final
    async def shutdown(self):
        if self.graceful_shutdown:
            self.info("shutting down gracefully")
            await self._shutdown_gracefully()
            self.info("graceful shut down complete")
        else:
            self.info("shutting down the hard way, task might not be re-queued...")

    def _create_worker_id(self) -> str:
        pid = os.getpid()
        threadid = threading.get_ident()
        hostname = socket.gethostname()
        # TODO: this might not be unique when using asyncio
        return f"{self._app.name}-worker-{hostname}-{pid}-{threadid}"

    def _clear_current(self):
        if self._timeout_callback_handle is not None:
            self._timeout_callback_handle.cancel()
            self._timeout_callback_handle = None
        self._current = None


def _retrieve_registered_task(
    task: Task,
    app: AsyncApp,
) -> RegisteredTask:
    registered = app.registry.get(task.name)
    if registered is None:
        available_tasks = list(app.registry)
        raise UnregisteredTask(task.name, available_tasks)
    return registered


async def task_wrapper(worker: Worker, task: Task) -> Task:
    worker.bind_worker_timeout()
    # Skips if already reserved
    if task.state is TaskState.CANCELLED:
        worker.info('Task(id="%s") already cancelled skipping it !', task.id)
        raise TaskAlreadyCancelled(task_id=task.id)
    # Parse task to retrieve recoverable errors and max retries
    task_fn, recoverable_errors = worker.parse_task(task)
    task_inputs = add_missing_args(task_fn, task.args)
    # Retry task until success, fatal error or max retry exceeded
    return await _retry_task(worker, task, task_fn, task_inputs, recoverable_errors)


async def _retry_task(
    worker: Worker,
    task: Task,
    task_fn: Callable,
    task_args: Dict,
    recoverable_errors: Tuple[Type[Exception], ...],
) -> Task:
    retries = task.retries_left
    if retries != task.max_retries:
        # In the case of the retry, let's reset the progress
        event = ProgressEvent.from_task(task=task)
        await worker.publish_event(event)
    try:
        task_res = task_fn(**task_args)
        if isawaitable(task_res):
            task_res = await task_res
    except recoverable_errors as e:
        # This will throw a MaxRetriesExceeded when necessary
        worker.check_retries(task, e)
        raise RecoverableError(e) from e
    update = TaskUpdate.done(datetime.now(timezone.utc))
    task = safe_copy(task, update=update.dict(exclude_unset=True))
    worker.info('Task(id="%s") complete, saving result...', task.id)
    async with worker.cancel_lock:
        await worker.publish_result_event(task_res, task)
    return task


def add_missing_args(fn: Callable, args: Dict[str, Any], **kwargs) -> Dict[str, Any]:
    # We make the choice not to raise in case of missing argument here, the error will
    # be correctly raise when the function is called
    from_kwargs = dict()
    sig = inspect.signature(fn)
    for param_name in sig.parameters:
        if param_name in args:
            continue
        kwargs_value = kwargs.get(param_name)
        if kwargs_value is not None:
            from_kwargs[param_name] = kwargs_value
    if from_kwargs:
        args = deepcopy(args)
        args.update(from_kwargs)
    return args


def _format_error(error: Exception) -> str:
    return "".join(traceback.format_exception(None, error, error.__traceback__))



"""
Simplified implementation of AllenNLP Registrable:
https://github.com/allenai/allennlp
"""

from __future__ import annotations

import logging
import os
from abc import ABC, abstractmethod
from collections import defaultdict
from pathlib import Path
from typing import (
    Callable,
    ClassVar,
    DefaultDict,
    Dict,
    List,
    Optional,
    Protocol,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

from pydantic import BaseModel, Field
from pydantic.parse import load_file

from icij_common.pydantic_utils import ICIJSettings
from icij_worker.utils.imports import VariableNotFound, import_variable

logger = logging.getLogger(__name__)

_RegistrableT = TypeVar("_RegistrableT", bound="Registrable")
_SubclassRegistry = Dict[str, _RegistrableT]
_SubclssNames = Dict[_SubclassRegistry, str]

T = TypeVar("T", bound="FromConfig")


class RegistrableMixin(ABC):
    _registry: ClassVar[DefaultDict[type, _SubclassRegistry]] = defaultdict(dict)
    _registered_names: ClassVar[DefaultDict[type, _SubclssNames]] = defaultdict(dict)

    default_implementation: Optional[str] = None

    @classmethod
    def register(
        cls, name: Optional[str] = None, exist_ok: bool = False
    ) -> Callable[[Type[RegistrableMixin]], Type[RegistrableMixin]]:
        # pylint: disable=protected-access
        registry = RegistrableMixin._registry[cls]
        registered_names = RegistrableMixin._registered_names[cls]

        def add_subclass_to_registry(
            subclass: Type[RegistrableMixin],
        ) -> Type[RegistrableMixin]:
            registered_name = name
            if registered_name is None:
                registered_key = subclass.registry_key.default
                if registered_key is None:
                    raise ValueError(
                        "no name provided and the class doesn't define a registry key"
                    )
                registered_name = getattr(subclass, registered_key).default

            if registered_name in registry:
                if exist_ok:
                    msg = (
                        f"{registered_name} has already been registered as "
                        f"{registry[registered_name].__name__}, but exist_ok=True, "
                        f"so overwriting with {cls.__name__}"
                    )
                    logger.info(msg)
                else:
                    msg = (
                        f"Cannot register {registered_name} as {cls.__name__}; "
                        f"name already in use for {registry[registered_name].__name__}"
                    )
                    raise ValueError(msg)
            registry[registered_name] = subclass
            registered_names[subclass] = registered_name
            return subclass

        return add_subclass_to_registry

    @classmethod
    def by_name(cls: Type[_RegistrableT], name: str) -> Callable[..., _RegistrableT]:
        logger.debug("instantiating registered subclass %s of %s", name, cls)
        subclass = cls.resolve_class_name(name)
        return cast(Type[_RegistrableT], subclass)

    @classmethod
    def resolve_class_name(cls: Type[_RegistrableT], name: str) -> Type[_RegistrableT]:
        # pylint: disable=protected-access
        sub_registry = RegistrableMixin._registry.get(cls, None)
        if sub_registry is None:
            for k, v in RegistrableMixin._registry.items():
                if issubclass(cls, k):
                    sub_registry = v
                    break
        if sub_registry is not None:
            subclass = sub_registry.get(name, None)
            if subclass is not None:
                return subclass
        if "." in name:
            try:
                subclass = import_variable(name)
            except ModuleNotFoundError as e:
                raise ValueError(
                    f"tried to interpret {name} as a path to a class "
                    f"but unable to import module {'.'.join(name.split('.')[:-1])}"
                ) from e
            except VariableNotFound as e:
                split = name.split(".")
                raise ValueError(
                    f"tried to interpret {name} as a path to a class "
                    f"but unable to find class {split[-1]} in {split[:-1]}"
                ) from e
            return subclass
        available = "\n-".join(cls.list_available())
        msg = f"""{name} is not a registered name for '{cls.__name__}'.
Available names are:
{available}

If your registered class comes from custom code, you'll need to import the\
 corresponding modules and use fully-qualified paths: "my_module.submodule.MyClass"
"""
        raise ValueError(msg)

    @classmethod
    def list_available(cls) -> List[str]:
        # pylint: disable=protected-access
        keys = list(RegistrableMixin._registry[cls].keys())
        return keys

    @classmethod
    @property
    def registered_name(cls) -> str:
        for (
            names
        ) in (
            RegistrableMixin._registered_names.values()
        ):  # pylint: disable=protected-access
            name = names.get(cls)
            if name is not None:
                return name
        raise ValueError("registration inconsistency")


class RegistrableConfig(BaseModel, RegistrableMixin):
    registry_key: ClassVar[str] = Field(const=True, default="name")

    @classmethod
    def parse_file(
        cls,
        path: Union[str, Path],
        *,
        content_type: str = None,
        encoding: str = "utf8",
        proto: Protocol = None,
        allow_pickle: bool = False,
    ) -> RegistrableConfig:
        obj = load_file(
            path,
            proto=proto,
            content_type=content_type,
            encoding=encoding,
            allow_pickle=allow_pickle,
            json_loads=cls.__config__.json_loads,
        )
        key = obj.pop(cls.registry_key.default)
        subcls = cls.resolve_class_name(key)
        return subcls(**obj)


class RegistrableSettings(RegistrableConfig, ICIJSettings):

    @classmethod
    def from_env(cls):
        key = cls.registry_key.default
        if cls.__config__.env_prefix is not None:
            key = cls.__config__.env_prefix + key
        registry_key = find_in_env(key, cls.__config__.case_sensitive)
        subcls = cls.resolve_class_name(registry_key)
        return subcls()


class FromConfig(ABC):
    @classmethod
    @abstractmethod
    def _from_config(cls, config: RegistrableConfig, **extras) -> FromConfig: ...


class RegistrableFromConfig(RegistrableMixin, FromConfig, ABC):
    @classmethod
    def from_config(cls, config: RegistrableConfig, **extras) -> FromConfig:
        name = getattr(config, config.registry_key.default).default
        subcls = cls.resolve_class_name(name)
        return subcls._from_config(config, **extras)  # pylint: disable=protected-access


def find_variable_loc_in_env(variable: str, case_sensitive: bool) -> Tuple[str, str]:
    if case_sensitive:
        try:
            return variable, os.environ[variable]
        except KeyError as e:
            raise ValueError(f"couldn't find {variable} in env variables") from e
    lowercase = variable.lower()
    for k, v in os.environ.items():
        if k.lower() == lowercase:
            return k, v
    raise ValueError(f"couldn't find {variable.upper()} in env variables")


def find_in_env(variable: str, case_sensitive: bool) -> str:
    return find_variable_loc_in_env(variable, case_sensitive)[1]



from pathlib import Path
from typing import Annotated, Optional

import typer

from icij_worker import AsyncApp, WorkerConfig
from icij_worker.backend import WorkerBackend, start_workers

_START_HELP = "Start a pool of workers running the provided app, reading the worker\
 configuration from the environment or an optionally provided file."
_APP_HELP = f'Path of the {AsyncApp.__name__} instance to run, fully qualified\
 ("module.sub.module.app_instance")'
_CONFIG_HELP = f"""Path to a serialized {WorkerConfig.__name__} JSON file.
By default, the configuration is read from the environment.
If present, file values will override environment variables values."""
_N_HELP = "Number of workers."
_BACKEND_HELP = "Python asynchronous backend used to create the worker pool."
_GROUP_HELP = """Worker group. The worker will only consume task coming from this
 group.
By default the worker will consume all tasks from the app group. 
"""
_DEFAULT_BACKEND = WorkerBackend.MULTIPROCESSING

worker_app = typer.Typer(name="workers")


@worker_app.command(help=_START_HELP)
def start(
    app: Annotated[str, typer.Argument(help=_APP_HELP)],
    config: Annotated[
        Optional[Path], typer.Option("-c", "--config", help=_CONFIG_HELP)
    ] = None,
    n: Annotated[int, typer.Option("--n-workers", "-n", help=_N_HELP)] = 1,
    backend: Annotated[
        WorkerBackend,
        typer.Option(
            help=_BACKEND_HELP,
            case_sensitive=False,
            show_default=_DEFAULT_BACKEND.value,
        ),
    ] = _DEFAULT_BACKEND,
    group: Annotated[
        Optional[str],
        typer.Option("--group", "-g", help=_GROUP_HELP),
    ] = None,
):
    start_workers(
        app=app, n_workers=n, config_path=config, backend=backend, group=group
    )



# pylint: disable=redefined-outer-name
# Test utils meant to be imported from clients libs to test their implem of workers
from __future__ import annotations

import asyncio
import json
import logging
import signal
import tempfile
from abc import ABC
from copy import deepcopy
from datetime import datetime, timezone
from pathlib import Path
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
    cast,
)

from pydantic import Field

import icij_worker
from icij_common.pydantic_utils import (
    ICIJModel,
    IgnoreExtraModel,
    jsonable_encoder,
    safe_copy,
)
from icij_common.test_utils import TEST_DB
from icij_worker import (
    AsyncApp,
    AsyncBackend,
    ManagerEvent,
    ResultEvent,
    RoutingStrategy,
    Task,
    TaskError,
    TaskState,
    Worker,
    WorkerConfig,
)
from icij_worker.app import AsyncAppConfig, TaskGroup
from icij_worker.event_publisher import EventPublisher
from icij_worker.exceptions import TaskQueueIsFull, UnknownTask
from icij_worker.objects import (
    CancelEvent,
    ErrorEvent,
    Message,
    TaskUpdate,
    WorkerEvent,
)
from icij_worker.task_manager import TaskManager, TaskManagerConfig
from icij_worker.typing_ import RateProgress
from icij_worker.utils.dependencies import DependencyInjectionError
from icij_worker.utils.logging_ import LogWithWorkerIDMixin

logger = logging.getLogger(__name__)

_has_pytest = False  # necessary because of the pytest decorators which requires pytest
# to be defined
try:
    import pytest

    _has_pytest = True
except ImportError:
    pass

if _has_pytest:

    # TODO: make this one a MockStorage
    class DBMixin(ABC):
        _task_collection = "tasks"
        _result_collection = "results"
        _error_collection = "errors"
        _lock_collection = "locks"
        _manager_event_collection = "manager_events"
        _worker_event_collection = "worker_events"

        _routing_strategy: RoutingStrategy

        def __init__(self, db_path: Path) -> None:
            self._db_path = db_path
            self._task_meta: Dict[str, Tuple[str, str]] = dict()

        @property
        def db_path(self) -> Path:
            return self._db_path

        def _write(self, data: Dict):
            self._db_path.write_text(json.dumps(jsonable_encoder(data)))

        def _read(self):
            return json.loads(self._db_path.read_text())

        @staticmethod
        def _task_key(task_id: str, db: Optional[str]) -> str:
            return str((task_id, db))

        def _get_task_db_name(self, task_id) -> str:
            if task_id not in self._task_meta:
                db = self._read()
                task_meta = dict()
                for k, v in db[self._task_collection].items():
                    (task_id, db) = eval(k)  # pylint: disable=eval-used
                    task_meta[task_id] = (db, v["group"])
                self._task_meta = task_meta
            try:
                return self._task_meta[task_id][0]
            except KeyError as e:
                raise UnknownTask(task_id) from e

        @classmethod
        def fresh_db(cls, db_path: Path):
            db = {
                cls._task_collection: dict(),
                cls._lock_collection: dict(),
                cls._error_collection: dict(),
                cls._result_collection: dict(),
                cls._manager_event_collection: dict(),
                cls._worker_event_collection: dict(),
            }
            db_path.write_text(json.dumps(db))

        async def get_task_group(self, task_id: str) -> Optional[str]:
            try:
                return self._task_meta[task_id][1]
            except KeyError as e:
                raise UnknownTask(task_id) from e

        async def save_result(self, result: ResultEvent):
            db = self._get_task_db_name(result.task_id)
            task_key = self._task_key(task_id=result.task_id, db=db)
            db = self._read()
            db[self._result_collection][task_key] = result
            self._write(db)

        async def save_error(self, error: ErrorEvent):
            db = self._get_task_db_name(task_id=error.task_id)
            task_key = self._task_key(task_id=error.task_id, db=db)
            db = self._read()
            errors = db[self._error_collection].get(task_key, [])
            errors.append(error)
            db[self._error_collection][task_key] = errors
            self._write(db)

        async def save_task_(self, task: Task, group: Optional[str]):
            new_task = False
            ns = None
            try:
                ns = await self.get_task_group(task_id=task.id)
            except UnknownTask:
                new_task = True
                if group is not None:
                    ns = group
                    db_name = self._routing_strategy.test_db(group)
                else:
                    db_name = TEST_DB
            else:
                if ns != group:
                    msg = (
                        f"DB task group ({ns}) differs from"
                        f" save task group: {group}"
                    )
                    raise ValueError(msg)
                db_name = self._get_task_db_name(task.id)
            update = TaskUpdate.from_task(task).dict(exclude_none=True, by_alias=True)
            update["group"] = group
            task_key = self._task_key(task_id=task.id, db=db_name)
            db = self._read()
            updated = db[self._task_collection].get(
                task_key, task.dict(exclude_none=True, by_alias=True)
            )
            updated.update(update)
            db[self._task_collection][task_key] = updated
            self._write(db)
            self._task_meta[task.id] = (db_name, ns)
            return new_task

        @staticmethod
        def _order_events(events: List[Dict]) -> float:
            events = [
                datetime.fromisoformat(evt["createdAt"]).timestamp() for evt in events
            ]
            return -min(events)

        async def _consume_(
            self,
            collection: str,
            sleep_interval: float,
            factory: Callable[[Any], R],
            group: Optional[str] = None,
            select: Optional[Callable[[R], bool]] = None,
            order: Optional[Callable[[R], Any]] = None,
        ) -> R:
            while "i'm waiting until I find something interesting":
                db = self._read()
                selected = deepcopy(db[collection])
                if collection == self._task_collection:
                    selected = {
                        k: v
                        for k, v in selected.items()
                        if k not in db[self._lock_collection]
                    }
                if group is not None:
                    selected = [
                        (k, t) for k, t in selected.items() if t.get("group") == group
                    ]
                    for k, t in selected:
                        t.pop("group", None)
                else:
                    selected = selected.items()
                selected = [(k, factory(t)) for k, t in selected]
                if select is not None:
                    selected = [(k, t) for k, t in selected if select(t)]
                if selected:
                    if order is not None:
                        k, t = min(selected, key=lambda x: order(x[1]))
                    else:
                        k, t = selected[0]
                    return t
                await asyncio.sleep(sleep_interval)

    @pytest.fixture(scope="session")
    def mock_db_session() -> Path:
        with tempfile.NamedTemporaryFile(prefix="mock-db", suffix=".json") as f:
            db_path = Path(f.name)
            DBMixin.fresh_db(db_path)
            yield db_path

    @pytest.fixture
    def mock_db(mock_db_session: Path) -> Path:
        # Wipe the DB
        DBMixin.fresh_db(mock_db_session)
        return mock_db_session

    class MockAppConfig(ICIJModel, LogWithWorkerIDMixin):
        # Just provide logging stuff to be able to see nice logs while doing TDD
        log_level: str = "DEBUG"
        loggers: List[str] = [icij_worker.__name__]

    _MOCKED_CONFIG: Optional[MockAppConfig] = None

    async def mock_async_config_enter(**_):
        global _MOCKED_CONFIG
        _MOCKED_CONFIG = MockAppConfig()
        logger.info("Loading mocked configuration %s", _MOCKED_CONFIG.json(indent=2))

    def lifespan_config() -> MockAppConfig:
        if _MOCKED_CONFIG is None:
            raise DependencyInjectionError("config")
        return _MOCKED_CONFIG

    def loggers_enter(worker_id: str, **_):
        config = lifespan_config()
        config.setup_loggers(worker_id=worker_id)
        logger.info("worker loggers ready to log 💬")

    mocked_app_deps = [
        ("configuration loading", mock_async_config_enter, None),
        ("loggers setup", loggers_enter, None),
    ]

    APP = AsyncApp(name="test-app", dependencies=mocked_app_deps)

    async def _hello_world(
        greeted: str, progress: Optional[RateProgress] = None
    ) -> str:
        if progress is not None:
            await progress(0.1)
        greeting = f"Hello {greeted} !"
        if progress is not None:
            await progress(0.99)
        return greeting

    @APP.task
    async def hello_world(greeted: str, progress: Optional[RateProgress] = None) -> str:
        return await _hello_world(greeted, progress)

    @APP.task(group="hello")
    async def grouped_hello_world(
        greeted: str, progress: Optional[RateProgress] = None
    ) -> str:
        return await _hello_world(greeted, progress)

    @APP.task
    def hello_world_sync(greeted: str) -> str:
        greeting = f"Hello {greeted} !"
        return greeting

    async def _sleep_for(
        duration: float, s: float = 0.01, progress: Optional[RateProgress] = None
    ):
        start = datetime.now()
        elapsed = 0
        while elapsed < duration:
            elapsed = (datetime.now() - start).total_seconds()
            await asyncio.sleep(s)
            if progress is not None:
                p = min(elapsed / duration, 1.0)
                await progress(p)

    @APP.task(max_retries=1)
    async def sleep_for(
        duration: float, s: float = 0.01, progress: Optional[RateProgress] = None
    ):
        await _sleep_for(duration, s, progress)

    short_timeout = 1
    short_tasks_group = TaskGroup(name="short", timeout_s=short_timeout)

    @APP.task(max_retries=3, group=short_tasks_group)
    async def sleep_for_short(
        duration: float, s: float = 0.01, progress: Optional[RateProgress] = None
    ):
        await _sleep_for(duration, s, progress)

    @APP.task(max_retries=666)
    async def often_retriable() -> str:
        pass

    @pytest.fixture(scope="session")
    def test_async_app() -> AsyncApp:
        return AsyncApp.load(f"{__name__}.APP")

    @pytest.fixture(scope="session")
    def app_config() -> AsyncAppConfig:
        return AsyncAppConfig()

    @pytest.fixture(scope="session")
    def late_ack_app_config() -> AsyncAppConfig:
        return AsyncAppConfig(late_ack=True)

    @pytest.fixture(scope="session")
    def test_async_app_late(late_ack_app_config: AsyncAppConfig) -> AsyncApp:
        return AsyncApp.load(f"{__name__}.APP", config=late_ack_app_config)

    class MockManagerConfig(TaskManagerConfig):
        backend: ClassVar[AsyncBackend] = Field(const=True, default=AsyncBackend.mock)
        db_path: Path
        event_refresh_interval_s: float = 0.1

    @TaskManager.register(AsyncBackend.mock)
    class MockManager(DBMixin, TaskManager):
        def __init__(
            self, app: AsyncApp, db_path: Path, event_refresh_interval_s: float = 0.1
        ):
            super().__init__(db_path)
            super(DBMixin, self).__init__(app)
            self._event_refresh_interval_s = event_refresh_interval_s

        @property
        def app(self) -> AsyncApp:
            return self._app

        async def _ensure_queue_size(self, task: Task):
            # pylint: disable=arguments-differ
            db_name = self._get_task_db_name(task.id)
            key = self._task_key(task.id, db=db_name)
            db = self._read()
            tasks = db[self._task_collection]
            n_queued = sum(
                1 for t in tasks.values() if t["state"] == TaskState.QUEUED.value
            )
            if (
                self.max_task_queue_size is not None
                and n_queued > self.max_task_queue_size
            ):
                raise TaskQueueIsFull(self.max_task_queue_size)
            db_task = tasks.get(key)
            if db_task is None:
                raise UnknownTask(task.id)

        async def _enqueue(self, task: Task):
            await self._ensure_queue_size(task)

        async def _requeue(self, task: Task):
            await self._ensure_queue_size(task)
            db_name = self._get_task_db_name(task.id)
            key = self._task_key(task.id, db=db_name)
            db = self._read()
            db_task = db[self._task_collection][key]
            db_task.pop("group")
            db_task = Task.parse_obj(db_task)
            update = TaskUpdate(progress=0.0, state=TaskState.QUEUED).dict(
                exclude_none=True, by_alias=True
            )
            requeued = safe_copy(db_task, update=update)
            db[self._task_collection][key] = requeued
            self._write(db)

        async def get_task(self, task_id: str) -> Task:
            db_name = self._get_task_db_name(task_id)
            key = self._task_key(task_id=task_id, db=db_name)
            db = self._read()
            try:
                tasks = db[self._task_collection]
            except KeyError as e:
                raise UnknownTask(task_id) from e
            task = deepcopy(tasks[key])
            task.pop("group")
            return Task.parse_obj(task)

        async def get_task_errors(self, task_id: str) -> List[ErrorEvent]:
            db = self._get_task_db_name(task_id)
            key = self._task_key(task_id=task_id, db=db)
            db = self._read()
            errors = db[self._error_collection]
            errors = errors.get(key, [])
            errors = [ErrorEvent.parse_obj(err) for err in errors]
            return errors

        async def get_task_result(self, task_id: str) -> ResultEvent:
            db = self._get_task_db_name(task_id)
            key = self._task_key(task_id=task_id, db=db)
            db = self._read()
            results = db[self._result_collection]
            try:
                return ResultEvent.parse_obj(results[key])
            except KeyError as e:
                raise UnknownTask(task_id) from e

        async def get_tasks(
            self,
            *,
            task_name: Optional[str] = None,
            state: Optional[Union[List[TaskState], TaskState]] = None,
            db: Optional[str] = None,
            **kwargs,
        ) -> List[Task]:
            # pylint: disable=arguments-differ
            db = self._read()
            tasks = db.values()
            if state:
                if isinstance(state, TaskState):
                    state = [state]
                state = set(state)
                tasks = (t for t in tasks if t.state in state)
            return list(tasks)

        async def _consume(self) -> ManagerEvent:
            events = await self._consume_(
                self._manager_event_collection,
                self._event_refresh_interval_s,
                list,
                select=lambda events: bool(  # pylint: disable=unnecessary-lambda
                    events
                ),
                order=self._order_events,
            )
            events = sorted(
                events,
                key=lambda e: datetime.fromisoformat(e["createdAt"]).timestamp(),
            )
            event = events[0]
            db = self._read()
            task_id = event["taskId"]
            db_name = self._get_task_db_name(task_id)
            key = self._task_key(task_id=task_id, db=db_name)
            db[self._manager_event_collection][key] = events[1:]
            self._write(db)
            return cast(ManagerEvent, Message.parse_obj(event))

        async def cancel(self, task_id: str, *, requeue: bool):
            cancel_event = CancelEvent(
                task_id=task_id, requeue=requeue, created_at=datetime.now(timezone.utc)
            )
            db_name = self._get_task_db_name(task_id)
            key = self._task_key(task_id=task_id, db=db_name)
            db = self._read()
            events = db[self._worker_event_collection]
            if key not in events:
                events[key] = []
            event_dict = cancel_event.dict(exclude_unset=True, by_alias=True)
            events[key].append(event_dict)
            self._write(db)

        @classmethod
        def _from_config(cls, config: MockManagerConfig, **extras) -> MockManager:
            tm = cls(
                config.app,
                config.db_path,
                event_refresh_interval_s=config.event_refresh_interval_s,
            )
            return tm

    R = TypeVar("R")

    class MockEventPublisher(DBMixin, EventPublisher):
        _excluded_from_event_update = {"error"}

        def __init__(self, db_path: Path):
            super().__init__(db_path)
            self.published_events = []

        async def _publish_event(self, event: ManagerEvent):
            db_name = self._get_task_db_name(event.task_id)
            key = self._task_key(task_id=event.task_id, db=db_name)
            db = self._read()
            event_dict = event.dict(exclude_unset=True, by_alias=True)
            if key not in db[self._manager_event_collection]:
                db[self._manager_event_collection][key] = []
            db[self._manager_event_collection][key].append(event_dict)
            self._write(db)
            self.published_events.append(event)

        def _get_db_task(self, db: Dict, *, task_id: str, db_name: str) -> Dict:
            tasks = db[self._task_collection]
            try:
                return tasks[self._task_key(task_id=task_id, db=db_name)]
            except KeyError as e:
                raise UnknownTask(task_id) from e

    @WorkerConfig.register()
    class MockWorkerConfig(WorkerConfig, IgnoreExtraModel):
        type: ClassVar[str] = Field(const=True, default=AsyncBackend.mock.value)

        db_path: Path
        log_level: str = "DEBUG"
        loggers: List[str] = [icij_worker.__name__]
        task_queue_poll_interval_s: float = 2.0

    @Worker.register(AsyncBackend.mock)
    class MockWorker(Worker, MockEventPublisher):
        def __init__(
            self,
            app: AsyncApp,
            worker_id: Optional[str] = None,
            *,
            group: Optional[str],
            db_path: Path,
            poll_interval_s: float,
            **kwargs,
        ):
            super().__init__(app, worker_id, group=group, **kwargs)
            MockEventPublisher.__init__(self, db_path)
            self._poll_interval_s = poll_interval_s
            self._worker_id = worker_id
            self._logger_ = logging.getLogger(__name__)
            self.terminated_cancelled_event_loop = False

        @property
        def app(self) -> AsyncApp:
            return self._app

        @property
        def watch_cancelled_task(self) -> Optional[asyncio.Task]:
            return self._watch_cancelled_task

        async def work_forever_async(self):
            await self._work_forever_async()

        @property
        def work_forever_task(self) -> Optional[asyncio.Task]:
            return self._work_forever_task

        @property
        def work_once_task(self) -> Optional[asyncio.Task]:
            return self._work_once_task

        @property
        def successful_exit(self) -> bool:
            return self._successful_exit

        @property
        def current(self) -> Optional[Task]:
            return self._current

        @current.setter
        def current(self, value: Optional[Task]):
            self._current = value

        async def signal_handler(self, signal_name: signal.Signals, *, graceful: bool):
            await self._signal_handler(signal_name, graceful=graceful)

        async def _aenter__(self):
            if not self._db_path.exists():
                raise OSError(f"worker DB was not initialized ({self._db_path})")

        @classmethod
        def _from_config(cls, config: MockWorkerConfig, **extras) -> MockWorker:
            worker = cls(
                db_path=config.db_path,
                poll_interval_s=config.task_queue_poll_interval_s,
                **extras,
            )
            return worker

        def _get_db_errors(self, task_id: str, db_name: str) -> List[TaskError]:
            key = self._task_key(task_id=task_id, db=db_name)
            db = self._read()
            errors = db[self._error_collection]
            try:
                return errors[key]
            except KeyError as e:
                raise UnknownTask(task_id) from e

        def _get_db_result(self, task_id: str, db_name: str) -> ResultEvent:
            key = self._task_key(task_id=task_id, db=db_name)
            db = self._read()
            try:
                errors = db[self._result_collection]
                return errors[key]
            except KeyError as e:
                raise UnknownTask(task_id) from e

        async def _acknowledge(self, task: Task):
            db_name = self._get_task_db_name(task.id)
            key = self._task_key(task.id, db_name)
            db = self._read()
            try:
                db[self._lock_collection].pop(key)
            except KeyError as e:
                raise UnknownTask(task.id) from e
            self._write(db)

        async def _negatively_acknowledge(self, nacked: Task):
            db_name = self._get_task_db_name(nacked.id)
            key = self._task_key(nacked.id, db_name)
            db = self._read()
            db[self._lock_collection].pop(key)
            self._write(db)

        @staticmethod
        def _task_factory(obj: Dict) -> Task:
            obj.pop("group", None)
            return Task.parse_obj(obj)

        async def _consume(self) -> Task:
            task = await self._consume_(
                self._task_collection,
                self._poll_interval_s,
                self._task_factory,
                group=self._group,
                select=lambda t: t.state is TaskState.QUEUED,
                order=lambda t: t.created_at,
            )
            db = self._read()
            db_name = self._get_task_db_name(task.id)
            key = self._task_key(task.id, db_name)
            db[self._lock_collection][key] = self._id
            self._write(db)
            return task

        async def _consume_worker_events(self) -> WorkerEvent:
            events = await self._consume_(
                self._worker_event_collection,
                self._poll_interval_s,
                list,
                group=self._group,
                select=lambda events: bool(  # pylint: disable=unnecessary-lambda
                    events
                ),
                order=self._order_events,
            )
            events = sorted(
                events,
                key=lambda e: datetime.fromisoformat(e["createdAt"]).timestamp(),
            )
            event = events[0]
            event = cast(WorkerEvent, Message.parse_obj(event))
            db = self._read()
            db_name = self._get_task_db_name(event.task_id)
            key = self._task_key(event.task_id, db_name)
            db[self._worker_event_collection][key] = events[1:]
            self._write(db)
            return event

        async def work_once(self):
            await self._work_once()

        async def publish_cancelled_event(self, cancel_event: CancelEvent):
            await self._publish_cancelled_event(cancel_event)



from abc import ABC, abstractmethod
from typing import Dict, List, Optional, Type

from icij_worker import RoutingStrategy, Task, TaskError, ResultEvent
from icij_worker.exceptions import UnknownTask
from icij_worker.objects import ErrorEvent, TaskUpdate
from icij_worker.task_storage import TaskStorage


class KeyValueStorage(TaskStorage, ABC):
    # Save each type in a different DB to speedup lookup, but that could be changed
    _tasks_db_name = "tasks"
    _results_db_name = "results"
    _errors_db_name = "errors"
    _routing_strategy: RoutingStrategy

    async def save_task_(self, task: Task, group: Optional[str]) -> bool:
        """When possible override this to be transactional"""
        key = self._key(task.id, obj_cls=ResultEvent)
        new_task = False
        try:
            ns = await self.get_task_group(task_id=task.id)
        except UnknownTask:
            new_task = True
            task = task.dict(exclude_unset=True)
            task["group"] = group
            await self._insert(self._tasks_db_name, task, key=key)
        else:
            if ns != group:
                msg = f"DB task group ({ns}) differs from" f" save task group: {group}"
                raise ValueError(msg)
            update = TaskUpdate.from_task(task).dict(exclude_none=True)
            await self._update(self._tasks_db_name, update, key=key)
        return new_task

    async def save_result(self, result: ResultEvent):
        res_key = self._key(result.task_id, obj_cls=ResultEvent)
        await self._insert(self._results_db_name, result.dict(), key=res_key)

    async def save_error(self, error: ErrorEvent):
        key = self._key(error.task_id, obj_cls=ErrorEvent)
        await self._add_to_array(self._errors_db_name, error.dict(), key=key)

    async def get_task(self, task_id: str) -> Task:
        key = self._key(task_id, obj_cls=Task)
        try:
            task = await self._read_key(self._tasks_db_name, key=key)
        except KeyError as e:
            raise UnknownTask(task_id) from e
        task.pop("group", None)
        return Task.parse_obj(task)

    async def get_task_group(self, task_id: str) -> Optional[str]:
        key = self._key(task_id, obj_cls=Task)
        try:
            task = await self._read_key(self._tasks_db_name, key=key)
        except KeyError as e:
            raise UnknownTask(task_id) from e
        group = task.get("group")
        return group

    async def get_task_errors(self, task_id: str) -> List[ErrorEvent]:
        key = self._key(task_id, obj_cls=TaskError)
        try:
            errors = await self._read_key(self._errors_db_name, key=key)
        except UnknownTask:
            return []
        errors = [TaskError.parse_obj(err) for err in errors]
        return errors

    async def get_task_result(self, task_id: str) -> ResultEvent:
        key = self._key(task_id, obj_cls=ResultEvent)
        try:
            result = await self._read_key(self._results_db_name, key=key)
        except KeyError as e:
            raise UnknownTask(task_id) from e
        return ResultEvent.parse_obj(result)

    @abstractmethod
    async def _read_key(self, db: str, *, key: str) -> Dict: ...

    @abstractmethod
    async def _insert(self, db: str, obj: Dict, *, key: str) -> str: ...

    @abstractmethod
    async def _update(self, db: str, update: Dict, *, key: str) -> str: ...

    @abstractmethod
    async def _add_to_array(self, db: str, obj: Dict, *, key: str) -> str: ...

    @abstractmethod
    def _key(self, task_id: str, obj_cls: Type) -> str: ...



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import List, Optional, Union

from icij_worker import ResultEvent, Task, TaskState
from icij_worker.objects import ErrorEvent, ProgressEvent
from icij_worker.routing_strategy import RoutingStrategy


class TaskStorageConfig(ABC):
    @abstractmethod
    def to_storage(self, routing_strategy: Optional[RoutingStrategy]) -> TaskStorage:
        pass


class TaskStorage(ABC):
    _routing_strategy: RoutingStrategy

    @abstractmethod
    async def get_task(self, task_id: str) -> Task: ...

    @abstractmethod
    async def get_task_errors(self, task_id: str) -> List[ErrorEvent]: ...

    @abstractmethod
    async def get_task_result(self, task_id: str) -> ResultEvent: ...

    @abstractmethod
    async def get_task_group(self, task_id: str) -> Optional[str]: ...

    @abstractmethod
    async def get_tasks(
        self,
        group: Optional[str],
        *,
        task_name: Optional[str] = None,
        state: Optional[Union[List[TaskState], TaskState]] = None,
        **kwargs,
    ) -> List[Task]: ...

    async def _save_progress_event(self, event: ProgressEvent):
        # Might be better to be overridden to be performed in a transactional manner
        # when possible
        task = await self.get_task(event.task_id)
        task = task.as_resolved(event)
        group = await self.get_task_group(event.task_id)
        await self.save_task_(task, group=group)

    @abstractmethod
    async def save_task_(self, task: Task, group: Optional[str]) -> bool: ...

    @abstractmethod
    async def save_result(self, result: ResultEvent): ...

    @abstractmethod
    async def save_error(self, error: ErrorEvent): ...



from __future__ import annotations

from contextlib import AbstractAsyncContextManager
from functools import lru_cache
from typing import ClassVar, Dict, Optional, cast

from aio_pika import RobustQueue
from aio_pika.abc import (
    AbstractIncomingMessage,
    AbstractQueueIterator,
    AbstractRobustConnection,
)
from pydantic import Field

from icij_common.pydantic_utils import safe_copy
from icij_worker import (
    AsyncApp,
    AsyncBackend,
    ManagerEvent,
    ResultEvent,
    Task,
    TaskError,
    Worker,
    WorkerConfig,
)
from icij_worker.event_publisher.amqp import (
    AMQPPublisher,
)
from icij_worker.objects import Message, TaskState, WorkerEvent
from icij_worker.routing_strategy import Routing
from icij_worker.utils.amqp import (
    AMQPConfigMixin,
    AMQPManagementClient,
    AMQPMixin,
    amqp_task_group_policy,
)

# TODO: remove this when project information is inside the tasks
_PROJECT_PLACEHOLDER = "placeholder_project_amqp"


@WorkerConfig.register()
class AMQPWorkerConfig(WorkerConfig, AMQPConfigMixin):
    type: ClassVar[str] = Field(const=True, default=AsyncBackend.amqp.value)


@Worker.register(AsyncBackend.amqp)
class AMQPWorker(Worker, AMQPMixin):

    def __init__(
        self,
        app: AsyncApp,
        management_client: AMQPManagementClient,
        worker_id: Optional[str] = None,
        *,
        group: Optional[str],
        broker_url: str,
        connection_timeout_s: float = 1.0,
        reconnection_wait_s: float = 5.0,
        inactive_after_s: Optional[float] = None,
        handle_signals: bool = True,
        teardown_dependencies: bool = False,
    ):
        super().__init__(
            app,
            worker_id,
            group=group,
            handle_signals=handle_signals,
            teardown_dependencies=teardown_dependencies,
        )
        AMQPMixin.__init__(
            self,
            broker_url,
            connection_timeout_s=connection_timeout_s,
            reconnection_wait_s=reconnection_wait_s,
            inactive_after_s=inactive_after_s,
        )
        self._management_client = management_client
        cancel_evt_queue_name = f"{self.worker_evt_routing().queue_name}-{self._id}"
        self._worker_evt_routing = safe_copy(
            self.worker_evt_routing(), update={"queue_name": cancel_evt_queue_name}
        )

        self._publisher: Optional[AMQPPublisher] = None
        self._connection_: Optional[AbstractRobustConnection] = None
        self._task_queue_: Optional[RobustQueue] = None
        self._task_queue_iterator: Optional[AbstractQueueIterator] = None
        self._worker_evt_queue_iterator: Optional[AbstractQueueIterator] = None
        self._task_routing: Optional[Routing] = None
        self._delivered: Dict[str, AbstractIncomingMessage] = dict()

        self._declare_exchanges = True

    @property
    def _app_id(self) -> str:
        return self._app.name

    async def _aenter__(self):
        await self._exit_stack.__aenter__()  # pylint: disable=unnecessary-dunder-call
        await self._exit_stack.enter_async_context(self._management_client)
        self._publisher = self._create_publisher()
        await self._exit_stack.enter_async_context(self._publisher)
        self._connection_ = self._publisher.connection
        self._channel_ = await self._connection.channel(
            publisher_confirms=True, on_return_raises=False
        )
        await self._channel.set_qos(prefetch_count=1, global_=False)
        await self._exit_stack.enter_async_context(self._channel)
        await self._bind_task_queue()
        await self._bind_worker_event_queue()

    async def _aexit__(self, exc_type, exc_val, exc_tb):
        event_transient_queue = await self._channel.get_queue(
            self._worker_evt_routing.queue_name
        )
        await event_transient_queue.delete(if_empty=False, if_unused=False)
        await self._exit_stack.__aexit__(exc_type, exc_val, exc_tb)

    async def _bind_task_queue(self):
        task_routing = self.task_routing(self._group)
        self._task_routing = task_routing
        self._task_queue_iterator, _, _ = await self._get_queue_iterator(
            self._task_routing,
            declare_exchanges=self._declare_exchanges,
            declare_queues=self._declare_exchanges,
        )
        self._task_queue_iterator = cast(
            AbstractAsyncContextManager, self._task_queue_iterator
        )
        await self._exit_stack.enter_async_context(self._task_queue_iterator)
        task_group = self._app.task_group(self._group)
        group_policy = amqp_task_group_policy(
            self._task_routing, task_group, self._app.config.max_task_queue_size
        )
        await self._management_client.set_policy(group_policy)

    async def _bind_worker_event_queue(self):
        self._worker_evt_queue_iterator, _, _ = await self._get_queue_iterator(
            self._worker_evt_routing,
            declare_queues=True,  # The worker always creates its own transient queue
            durable_queues=False,
            declare_exchanges=self._declare_exchanges,
        )
        self._worker_evt_queue_iterator = cast(
            AbstractAsyncContextManager, self._worker_evt_queue_iterator
        )
        await self._exit_stack.enter_async_context(self._worker_evt_queue_iterator)

    async def _consume(self) -> Task:
        # TODO: remove this now that task groups exist
        while "I'm waiting to get a known task":
            # pylint: disable=unnecessary-dunder-call
            message: AbstractIncomingMessage = await self._task_messages_it.__anext__()
            task = Task.parse_raw(message.body)
            self._delivered[task.id] = message
            # This behavior is not shared with other implems, it's AMQP specific and
            # avoid a worker consuming in loops tasks which it can't handle.
            # This also bypasses the normal/tested/share7 behavior of the Worker class
            if not await self._ensure_can_consume(task):
                continue
            return task

    async def _consume_worker_events(self) -> WorkerEvent:
        # pylint: disable=unnecessary-dunder-call
        message: AbstractIncomingMessage = await self._worker_events_it.__anext__()
        await message.ack()
        event = cast(WorkerEvent, Message.parse_raw(message.body))
        return event

    @property
    def _task_messages_it(self) -> AbstractQueueIterator:
        if self._task_queue_iterator is None:
            msg = (
                f"Worker as no task message iterator, "
                f"please call {AMQPWorker.__aenter__} first"
            )
            raise ValueError(msg)
        return self._task_queue_iterator

    @property
    def _worker_events_it(self) -> AbstractQueueIterator:
        if self._worker_evt_queue_iterator is None:
            msg = (
                f"Worker as no cancel events iterator, "
                f"please call {AMQPWorker.__aenter__} first"
            )
            raise ValueError(msg)
        return self._worker_evt_queue_iterator

    async def _acknowledge(self, task: Task):
        message = self._delivered[task.id]
        await message.ack()

    async def _negatively_acknowledge(self, nacked: Task):
        # pylint: disable=unused-argument
        message = self._delivered[nacked.id]
        requeue = nacked.state is not TaskState.ERROR
        await message.nack(requeue=requeue)

    async def _publish_event(self, event: ManagerEvent):
        await self._publisher.publish_event(event)

    async def _save_result(self, result: ResultEvent):
        await self._publisher.publish_result(result)

    async def _save_error(self, error: TaskError):
        await self._publisher.publish_error(error)

    @lru_cache()
    def task_routing(self, group: Optional[str]) -> Routing:
        routing = self._routing_strategy.amqp_task_routing(group)
        return routing

    def _create_publisher(self) -> AMQPPublisher:
        return AMQPPublisher(
            self._logger,
            broker_url=self._broker_url,
            connection_timeout_s=self._reconnection_wait_s,
            reconnection_wait_s=self._reconnection_wait_s,
            app_id=self._app.name,
        )

    @classmethod
    def _from_config(cls, config: AMQPWorkerConfig, **extras) -> AMQPWorker:
        management_client = config.to_management_client()
        worker = cls(
            management_client=management_client,
            broker_url=config.broker_url,
            connection_timeout_s=config.connection_timeout_s,
            reconnection_wait_s=config.reconnection_wait_s,
            inactive_after_s=config.inactive_after_s,
            **extras,
        )
        return worker

    async def _ensure_can_consume(self, task: Task) -> bool:
        if task.name not in self._app.registry:
            msg = (
                f'Task(id="%s") has unknown name "{task.name}" for worker. '
                f"Nacking it with requeue."
            )
            self.error(msg)
            task_msg = self._delivered[task.id]
            await task_msg.nack(requeue=True)
            return False
        return True



# AMQP
from icij_common.pydantic_utils import to_lower_camel

AMQP_MANAGER_EVENTS_X = "exchangeManagerEvents"
AMQP_MANAGER_EVENTS_QUEUE = "MANAGER_EVENT"
AMQP_MANAGER_EVENTS_ROUTING_KEY = "routingKeyManagerEvents"

AMQP_MANAGER_EVENTS_DL_X = "exchangeDLQManagerEvents"
AMQP_MANAGER_EVENTS_DL_QUEUE = "MANAGER_EVENT_DLQ"
AMQP_MANAGER_EVENTS_DL_ROUTING_KEY = "routingKeyDLQManagerEvents"

AMQP_TASKS_DL_X = "exchangeDLQTasks"
AMQP_TASKS_DL_QUEUE = "TASK_DLQ"
AMQP_TASKS_DL_ROUTING_KEY = "routingKeyDLQTasks"

AMQP_TASKS_X = "exchangeMainTasks"
AMQP_TASKS_QUEUE = "TASK"
AMQP_TASKS_ROUTING_KEY = "routingKeyMainTasks"

AMQP_WORKER_EVENTS_X = "exchangeWorkerEvents"
AMQP_WORKER_EVENTS_QUEUE = "WORKER_EVENT"
AMQP_WORKER_EVENTS_ROUTING_KEY = "routingKeyWorkerEvents"

AMQP_TASK_QUEUE_PRIORITY = 1000

_CREATED_AT = "created_at"
_TASK_ID = "task_id"

# General
TASK_ARGUMENTS_DEPRECATED = "arguments"
TASK_ARGS = "args"
TASK_CANCELLED_AT = "cancelled_at"
TASK_COMPLETED_AT = "completed_at"
TASK_CREATED_AT = _CREATED_AT
TASK_GROUP = "group"
TASK_ID = "id"
TASK_MAX_RETRIES = "max_retries"
TASK_NAME = "name"
TASK_NAMESPACE_DEPRECATED = "namespace"
TASK_PROGRESS = "progress"
TASK_RETRIES_LEFT = "retries_left"
TASK_STATE = "state"

TASK_ERRORS_TASK_ID = _TASK_ID

TASK_RESULT_CREATED_AT = _CREATED_AT
TASK_RESULT_TASK_ID = _TASK_ID
TASK_RESULT_RESULT = "result"

# POSTGRES
POSTGRES_TASKS_TABLE = "tasks"
POSTGRES_TASK_ERRORS_TABLE = "errors"
POSTGRES_TASK_RESULTS_TABLE = "results"

POSTGRES_TASK_DB_NAME = "name"
POSTGRES_TASK_DB_IS_LOCKED = "is_locked"
POSTGRES_TASK_DBS_TABLE = "task_dbs"

# Neo4j
NEO4J_TASK_NODE = "_Task"
NEO4J_TASK_ARGS = to_lower_camel(TASK_ARGS)
NEO4J_TASK_ARGUMENTS_DEPRECATED = to_lower_camel(TASK_ARGUMENTS_DEPRECATED)
NEO4J_TASK_CANCELLED_AT = to_lower_camel(TASK_CANCELLED_AT)
NEO4J_TASK_COMPLETED_AT = to_lower_camel(TASK_COMPLETED_AT)
NEO4J_TASK_CREATED_AT = to_lower_camel(TASK_CREATED_AT)
NEO4J_TASK_ID = to_lower_camel(TASK_ID)
NEO4J_TASK_INPUTS_DEPRECATED = "inputs"
NEO4J_TASK_GROUP = to_lower_camel(TASK_GROUP)
NEO4J_TASK_MAX_RETRIES = to_lower_camel(TASK_MAX_RETRIES)
NEO4J_TASK_NAME = to_lower_camel(TASK_NAME)
NEO4J_TASK_NAMESPACE_DEPRECATED = to_lower_camel(TASK_NAMESPACE_DEPRECATED)
NEO4J_TASK_PROGRESS = to_lower_camel(TASK_PROGRESS)
NEO4J_TASK_RETRIES_DEPRECATED = "retries"
NEO4J_TASK_RETRIES_LEFT = to_lower_camel(TASK_RETRIES_LEFT)
NEO4J_TASK_TYPE_DEPRECATED = "type"

NEO4J_TASK_CANCEL_EVENT_NODE = "_CancelEvent"
NEO4J_TASK_CANCEL_EVENT_CREATED_AT_DEPRECATED = "createdAt"
NEO4J_TASK_CANCEL_EVENT_CANCELLED_AT = "cancelledAt"
NEO4J_TASK_CANCEL_EVENT_EFFECTIVE = "effective"
NEO4J_TASK_CANCEL_EVENT_REQUEUE = "requeue"
NEO4J_TASK_CANCELLED_BY_EVENT_REL = "_CANCELLED_BY"

NEO4J_TASK_MANAGER_EVENT_NODE = "_ManagerEvent"
NEO4J_TASK_MANAGER_EVENT_NODE_CREATED_AT = to_lower_camel(_CREATED_AT)
NEO4J_TASK_MANAGER_EVENT_EVENT = "event"

NEO4J_TASK_LOCK_NODE = "_TaskLock"
NEO4J_TASK_LOCK_TASK_ID = to_lower_camel(_TASK_ID)
NEO4J_TASK_LOCK_WORKER_ID = "workerId"

NEO4J_TASK_ERROR_NODE = "_TaskError"
NEO4J_TASK_ERROR_DETAIL_DEPRECATED = "detail"  # use stacktrace
NEO4J_TASK_ERROR_ID_DEPRECATED = "id"
NEO4J_TASK_ERROR_MESSAGE = "message"
NEO4J_TASK_ERROR_NAME = "name"
NEO4J_TASK_ERROR_OCCURRED_AT_DEPRECATED = "occurredAt"
NEO4J_TASK_ERROR_STACKTRACE = "stacktrace"
NEO4J_TASK_ERROR_TITLE_DEPRECATED = "title"  # use message

NEO4J_TASK_ERROR_OCCURRED_TYPE = "_OCCURRED_DURING"
NEO4J_TASK_ERROR_OCCURRED_TYPE_OCCURRED_AT = "occurredAt"
NEO4J_TASK_ERROR_OCCURRED_TYPE_RETRIES_LEFT = "retriesLeft"

NEO4J_TASK_RESULT_NODE = "_TaskResult"
NEO4J_TASK_HAS_RESULT_TYPE = "_HAS_RESULT"
NEO4J_TASK_RESULT_RESULT = "result"



import functools
import logging
import multiprocessing
import signal
from concurrent.futures import (
    CancelledError,
    Future,
    InvalidStateError,
    ProcessPoolExecutor,
    as_completed,
)
from contextlib import contextmanager
from typing import Callable, Dict, List, Optional, Set, Tuple

import sys

import icij_worker
from icij_common.logging_utils import setup_loggers
from icij_worker import AsyncApp, Worker, WorkerConfig

logger = logging.getLogger(__name__)

_HANDLED_SIGNALS = [signal.SIGTERM, signal.SIGINT]
if sys.platform == "win32":
    _HANDLED_SIGNALS += [signal.CTRL_C_EVENT, signal.CTRL_BREAK_EVENT]


def _mp_work_forever(
    app: str,
    config: WorkerConfig,
    *,
    worker_group: Optional[str],
    worker_extras: Optional[Dict] = None,
    app_deps_extras: Optional[Dict] = None,
):
    setup_loggers(["__main__", icij_worker.__name__])
    try:
        if worker_extras is None:
            worker_extras = dict()
        if app_deps_extras is None:
            app_deps_extras = dict()
        # For multiprocessing, lifespan dependencies need to be run once per process
        app = AsyncApp.load(app)
        app.filter_tasks(worker_group)
        worker = Worker.from_config(
            config,
            app=app,
            group=worker_group,
            **worker_extras,
        )
        deps_cm = app.lifetime_dependencies(
            worker_id=worker.id, worker_config=config, **app_deps_extras
        )
        # From now on, the deps_cm should have setup loggers, we can let it log errors,
        # we get out of here
        # This is ugly, but we have to work around the fact that we can't use asyncio
        # code here
        worker.loop.run_until_complete(
            deps_cm.__aenter__()  # pylint: disable=unnecessary-dunder-call
        )
    except BaseException as e:
        msg = "Error occurred during app loading or dependency injection: %s"
        logger.exception(msg, e)
        raise e
    try:
        worker.work_forever()
    finally:
        worker.info("worker stopped working, tearing down %s dependencies", app.name)
        worker.loop.run_until_complete(deps_cm.__aexit__(*sys.exc_info()))
        worker.info("dependencies closed for %s !", app.name)


def signal_handler(sig_num, *_):
    logger.exception(
        "received %s, triggering process executor shutdown !",
        signal.Signals(sig_num).name,
    )


def setup_main_process_signal_handlers():
    for s in _HANDLED_SIGNALS:
        signal.signal(s, signal_handler)


def _get_mp_async_runner(
    app: str,
    config: WorkerConfig,
    n_workers: int,
    *,
    worker_extras: Optional[Dict] = None,
    app_deps_extras: Optional[Dict] = None,
    group: Optional[str],
) -> Tuple[ProcessPoolExecutor, List[Callable]]:
    # This function is here to avoid code duplication, it will be removed

    # Here we set maxtasksperchild to 1. Each worker has a single never ending task
    # which consists in working forever. Additionally, in some cases using
    # maxtasksperchild=1 seems to help to terminate the worker pull
    # (cpython bug: https://github.com/python/cpython/pull/8009)
    mp_ctx = multiprocessing.get_context("spawn")
    executor = ProcessPoolExecutor(max_workers=n_workers, mp_context=mp_ctx)
    kwds = {
        "app": app,
        "config": config,
        "worker_extras": worker_extras,
        "worker_group": group,
        "app_deps_extras": app_deps_extras,
    }
    futures = []
    for _ in range(n_workers):
        futures.append(functools.partial(executor.submit, _mp_work_forever, **kwds))
    return executor, futures


def _cancel_other_callback(errored: Future, others: List[Future]):
    try:
        e = errored.exception()
    except CancelledError:
        return
    if isinstance(e, CancelledError):
        return
    for task in others:
        if task is errored:
            continue
        try:
            task.cancel(f"cancelled by {errored}")
        except InvalidStateError:
            pass


@contextmanager
def _handle_executor_termination(
    executor: ProcessPoolExecutor, futures: Set[Future], handle_signals: bool
):
    try:
        yield
    except KeyboardInterrupt as e:
        if not handle_signals:
            logger.info(
                "received %s, triggering process executor worker shutdown !",
                KeyboardInterrupt.__name__,
            )
        else:
            msg = (
                f"Received {KeyboardInterrupt.__name__} while SIGINT was expected to"
                f" be handled"
            )
            raise RuntimeError(msg) from e
    finally:
        msg = "Worker terminated by the executor"
        exc = CancelledError(msg)
        for f in futures:
            if not f.done():
                f.set_exception(exc)
        logger.info("Sending termination signal to workers (SIGTERM)...")
        executor.shutdown(wait=False, cancel_futures=True)
        logger.info("Terminated worker executor !")


@contextmanager
def run_workers_with_multiprocessing_cm(
    app: str,
    n_workers: int,
    config: WorkerConfig,
    *,
    worker_extras: Optional[Dict] = None,
    app_deps_extras: Optional[Dict] = None,
    group: Optional[str],
):
    if n_workers < 1:
        raise ValueError("n_workers must be >=1")
    logger.info("Creating multiprocessing executor with %s workers", n_workers)
    executor, worker_runners = _get_mp_async_runner(
        app,
        config,
        n_workers,
        worker_extras=worker_extras,
        app_deps_extras=app_deps_extras,
        group=group,
    )
    futures = set()
    for process_runner in worker_runners:
        future = process_runner()
        futures.add(future)
    for f in futures:
        f.add_done_callback(functools.partial(_cancel_other_callback, others=futures))
    logger.info("started %s workers for app %s", n_workers, app)
    original_error = None
    with _handle_executor_termination(executor, futures, True):
        for f in as_completed(futures):
            try:
                f.result()
            except CancelledError:
                pass
            except Exception as e:  # pylint: disable=broad-exception-caught
                original_error = e
        del futures
    if original_error:
        raise original_error


def run_workers_with_multiprocessing(
    app: str,
    n_workers: int,
    config: WorkerConfig,
    *,
    worker_extras: Optional[Dict] = None,
    app_deps_extras: Optional[Dict] = None,
    group: Optional[str],
):
    if n_workers < 1:
        raise ValueError("n_workers must be >=1")
    logger.info("Creating multiprocessing executor with %s workers", n_workers)
    executor, worker_runners = _get_mp_async_runner(
        app,
        config,
        n_workers,
        worker_extras=worker_extras,
        app_deps_extras=app_deps_extras,
        group=group,
    )
    setup_main_process_signal_handlers()
    futures = set()
    for process_runner in worker_runners:
        future = process_runner()
        futures.add(future)
    for f in futures:
        f.add_done_callback(functools.partial(_cancel_other_callback, others=futures))
    logger.info("started %s workers for app %s", n_workers, app)
    original_error = None
    with _handle_executor_termination(executor, futures, True):
        for f in as_completed(futures):
            try:
                f.result()
            except CancelledError:
                pass
            except Exception as e:  # pylint: disable=broad-exception-caught
                original_error = e
        del futures
    if original_error:
        raise original_error



from __future__ import annotations

import functools
from contextlib import AsyncExitStack
from pathlib import Path
from typing import Dict, List, Optional, Type, Union

import ujson
from sqlitedict import SqliteDict

from icij_common.pydantic_utils import ICIJModel
from icij_worker import RoutingStrategy, Task, TaskState
from icij_worker.exceptions import UnknownTask
from icij_worker.task_storage import TaskStorageConfig
from icij_worker.task_storage.key_value import KeyValueStorage


class FSKeyValueStorageConfig(ICIJModel, TaskStorageConfig):
    db_path: Path

    def to_storage(
        self, routing_strategy: Optional[RoutingStrategy]
    ) -> FSKeyValueStorage:
        return FSKeyValueStorage(self.db_path, routing_strategy)


class FSKeyValueStorage(KeyValueStorage):
    # Save each type in a different DB to speedup lookup
    _tasks_db_name = "tasks"
    _results_db_name = "results"
    _errors_db_name = "errors"
    # pylint: disable=c-extension-no-member
    _encode = functools.partial(ujson.encode, default=str)
    _decode = functools.partial(ujson.decode)

    # pylint: enable=c-extension-no-member

    def __init__(
        self, db_path: Path, routing_strategy: Optional[RoutingStrategy] = None
    ):
        super().__init__()
        if routing_strategy is None:
            routing_strategy = RoutingStrategy()
        self._routing_strategy = routing_strategy
        self._db_path = str(db_path)
        self._exit_stack = AsyncExitStack()
        self._dbs = dict()
        # TODO: add support for 1 DB / group
        self._dbs = self._make_ns_dbs()

    async def __aenter__(self):
        for db in self._dbs.values():
            self._exit_stack.enter_context(db)

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._exit_stack.__aexit__(exc_type, exc_val, exc_tb)

    async def _read_key(self, db: str, *, key: str):
        db = self._dbs[db]
        try:
            return db[key]
        except KeyError as e:
            raise UnknownTask(key) from e

    async def _insert(self, db: str, obj: Dict, *, key: str):
        db = self._dbs[db]
        try:
            db[key] = obj
        except KeyError as e:
            raise UnknownTask(key) from e
        db.commit()

    async def _update(self, db: str, update: Dict, *, key: str):
        db = self._dbs[db]
        try:
            task = db[key]
        except KeyError as e:
            raise UnknownTask(key) from e
        task.update(update)
        db[key] = task
        db.commit()

    async def _add_to_array(self, db: str, obj: Dict, *, key: str):
        db = self._dbs[db]
        values = obj.get(key, [])
        values.append(obj)
        db[key] = values
        db.commit()

    def _key(self, task_id: str, obj_cls: Type) -> str:
        # Since each object type is saved in a different DB, we index by task ID
        return task_id

    async def get_tasks(
        self,
        group: Optional[str],
        *,
        task_name: Optional[str] = None,
        state: Optional[Union[List[TaskState], TaskState]] = None,
        **kwargs,
    ) -> List[Task]:
        states = set()
        if state is not None:
            if isinstance(state, TaskState):
                states = {state}
            states = set(states)
        tasks = self._dbs[self._tasks_db_name].values()
        if group is not None:
            tasks = (t for t in tasks if t.get("group") == group)
        if task_name is not None:
            tasks = (t for t in tasks if t["name"] == task_name)
        if states:
            tasks = (t for t in tasks if t["state"] in states)
        tasks = list(tasks)
        for t in tasks:
            t.pop("group", None)
        task = [Task.parse_obj(t) for t in tasks]
        return task

    def _make_db(self, filename: str, *, name: str) -> SqliteDict:
        return SqliteDict(
            filename,
            tablename=name,
            encode=self._encode,
            decode=self._decode,
            journal_mode="DEFAULT",
        )

    def _make_ns_dbs(self) -> Dict:
        return {
            self._tasks_db_name: self._make_db(self._db_path, name=self._tasks_db_name),
            self._results_db_name: self._make_db(
                self._db_path, name=self._results_db_name
            ),
            self._errors_db_name: self._make_db(
                self._db_path, name=self._errors_db_name
            ),
        }



from __future__ import annotations

import json
import logging
import traceback
from abc import ABC, abstractmethod
from datetime import datetime, timezone
from enum import Enum, unique
from functools import lru_cache
from typing import Callable, ClassVar, Sequence, Union, cast

from pydantic import Field, root_validator, validator
from pydantic.utils import ROOT_KEY
from typing_extensions import Any, Dict, List, Optional, final

from icij_common import neo4j
from icij_common.pydantic_utils import (
    ICIJModel,
    ISODatetime,
    LowerCamelCaseModel,
    NoEnumModel,
    safe_copy,
)
from icij_worker.constants import (
    NEO4J_TASK_CANCEL_EVENT_CANCELLED_AT,
    NEO4J_TASK_CANCEL_EVENT_REQUEUE,
    NEO4J_TASK_COMPLETED_AT,
    NEO4J_TASK_ERROR_MESSAGE,
    NEO4J_TASK_ERROR_NAME,
    NEO4J_TASK_ERROR_OCCURRED_TYPE_OCCURRED_AT,
    NEO4J_TASK_ERROR_OCCURRED_TYPE_RETRIES_LEFT,
    NEO4J_TASK_ERROR_STACKTRACE,
    NEO4J_TASK_ID,
    NEO4J_TASK_NODE,
    NEO4J_TASK_RESULT_RESULT,
)
from icij_worker.typing_ import AbstractSetIntStr, DictStrAny, MappingIntStrAny
from icij_worker.utils.registrable import RegistrableMixin

logger = logging.getLogger(__name__)

PROGRESS_HANDLER_ARG = "progress_handler"
_TASK_SCHEMA = None

TASK_ERROR_CAUSE = "cause"


@unique
class AsyncBackend(str, Enum):
    # pylint: disable=invalid-name@
    mock = "mock"
    neo4j = "neo4j"
    amqp = "amqp"


class FromTask(ABC):

    @classmethod
    @abstractmethod
    def from_task(cls, task: Task, **kwargs) -> FromTask: ...


@unique
class TaskState(str, Enum):
    CREATED = "CREATED"
    QUEUED = "QUEUED"
    RUNNING = "RUNNING"
    ERROR = "ERROR"
    DONE = "DONE"
    CANCELLED = "CANCELLED"

    @classmethod
    def resolve_update_state(cls, stored: Task, update: TaskUpdate) -> TaskState:
        # A done task is always done
        if stored.state is TaskState.DONE:
            return stored.state
        # A task store as ready can't be updated unless there's a new ready state
        # (for instance ERROR -> DONE)
        if stored.state in READY_STATES and update.state not in READY_STATES:
            return stored.state
        if update.state is TaskState.QUEUED and stored.state is TaskState.RUNNING:
            # We have to store the most recent state
            if update.cancelled_at is not None:
                if (
                    stored.cancelled_at is None
                    or stored.cancelled_at < update.cancelled_at
                ):
                    return update.state
                return stored.state
            return stored.state
        # Otherwise the true state is the most advanced on in the state machine
        return max(stored.state, update.state)

    def __gt__(self, other: TaskState) -> bool:
        return state_precedence(self) < state_precedence(other)

    def __ge__(self, other: TaskState) -> bool:
        return state_precedence(self) <= state_precedence(other)

    def __lt__(self, other: TaskState) -> bool:
        return state_precedence(self) > state_precedence(other)

    def __le__(self, other: TaskState) -> bool:
        return state_precedence(self) >= state_precedence(other)


READY_STATES = frozenset({TaskState.DONE, TaskState.ERROR, TaskState.CANCELLED})
# Greatly inspired from Celery
PRECEDENCE = [
    TaskState.DONE,
    TaskState.ERROR,
    TaskState.CANCELLED,
    TaskState.RUNNING,
    TaskState.QUEUED,
    TaskState.CREATED,
]
PRECEDENCE_LOOKUP = dict(zip(PRECEDENCE, range(len(PRECEDENCE))))


def state_precedence(state: TaskState) -> int:
    return PRECEDENCE_LOOKUP[state]


class Neo4jDatetimeMixin(ISODatetime):
    @classmethod
    def _validate_neo4j_datetime(cls, value: Any) -> datetime:
        # Trick to avoid having to import neo4j here
        if not isinstance(value, datetime) and hasattr(value, "to_native"):
            value = value.to_native()
        return value


def _encode_registrable(v, **kwargs):
    return json.dumps(v.dict(exclude_unset=True), **kwargs)


class Registrable(ICIJModel, RegistrableMixin, ABC):
    registry_key: ClassVar[str] = Field(const=True, default="@type")

    class Config:
        json_encoders = {"Registrable": _encode_registrable}

    @root_validator(pre=True)
    def validate_type(cls, values):  # pylint: disable=no-self-argument
        values.pop("@type", None)
        return values

    @classmethod
    def parse_obj(cls, obj: Dict) -> Registrable:
        key = obj.pop(cls.registry_key.default)
        subcls = cls.resolve_class_name(key)
        return subcls(**obj)

    def dict(
        self,
        *,
        include: Optional[Union[AbstractSetIntStr, MappingIntStrAny]] = None,
        exclude: Optional[Union[AbstractSetIntStr, MappingIntStrAny]] = None,
        by_alias: bool = False,
        skip_defaults: Optional[bool] = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
    ) -> DictStrAny:
        as_dict = super().dict(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            skip_defaults=skip_defaults,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )
        if not exclude or self.__class__.registry_key.default not in exclude:
            as_dict[self.__class__.registry_key.default] = (
                self.__class__.registered_name
            )
        return as_dict

    def json(
        self,
        *,
        include: Optional[Union["AbstractSetIntStr", "MappingIntStrAny"]] = None,
        exclude: Optional[Union["AbstractSetIntStr", "MappingIntStrAny"]] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        encoder: Optional[Callable[[Any], Any]] = None,
        models_as_dict: bool = True,
        **dumps_kwargs: Any,
    ) -> str:
        encoder = cast(Callable[[Any], Any], encoder or self.__json_encoder__)
        data = dict(
            self._iter(
                to_dict=models_as_dict,
                by_alias=by_alias,
                include=include,
                exclude=exclude,
                exclude_unset=exclude_unset,
                exclude_defaults=exclude_defaults,
                exclude_none=exclude_none,
            )
        )
        if self.__custom_root_type__:
            data = data[ROOT_KEY]
        data[self.__class__.registry_key.default] = self.__class__.registered_name
        return self.__config__.json_dumps(data, default=encoder, **dumps_kwargs)


class Message(Registrable): ...  # pylint: disable=multiple-statements


@Message.register("Task")
class Task(Message, NoEnumModel, LowerCamelCaseModel, Neo4jDatetimeMixin):
    id: str
    name: str
    args: Optional[Dict[str, object]] = None
    state: TaskState
    progress: Optional[float] = None
    created_at: datetime
    completed_at: Optional[datetime] = None
    cancelled_at: Optional[datetime] = None
    retries_left: Optional[int] = None
    max_retries: Optional[int] = None

    _non_inherited_from_event = [
        "requeue",
        "created_at",
        "error",
        "occurred_at",
        "task_name",
        "result",
        "task_id",
        "retries_left",
    ]

    @validator("args", pre=True, always=True)
    def args_as_dict(cls, v: Optional[Dict[str, Any]]):
        # pylint: disable=no-self-argument
        if v is None:
            v = dict()
        return v

    @root_validator(pre=True)
    def retries_left_should_default_to_max_retries_when_missing(
        cls, values: Dict[str, Any]
    ) -> Dict[str, Any]:
        # pylint: disable=no-self-argument
        max_retries = values.get("max_retries")
        if values.get("retries_left") is None and max_retries is not None:
            values["retries_left"] = max_retries
        return values

    @classmethod
    def create(cls, *, task_id: str, task_name: str, args: Dict[str, Any]) -> Task:
        created_at = datetime.now(timezone.utc)
        state = TaskState.CREATED
        return cls(
            id=task_id,
            name=task_name,
            args=args,
            created_at=created_at,
            state=state,
        )

    def with_max_retries(self, max_retries: Optional[int]) -> Task:
        as_dict = self.dict()
        as_dict.pop("max_retries", None)
        return Task(max_retries=max_retries, **as_dict)

    @validator("args", pre=True)
    def _validate_args(cls, value: Any):  # pylint: disable=no-self-argument
        if isinstance(value, str):
            value = json.loads(value)
        return value

    @validator("created_at", pre=True)
    def _validate_created_at(cls, value: Any):  # pylint: disable=no-self-argument
        return cls._validate_neo4j_datetime(value)

    @validator("completed_at", pre=True)
    def _validate_completed_at(cls, value: Any):  # pylint: disable=no-self-argument
        return cls._validate_neo4j_datetime(value)

    @validator("cancelled_at", pre=True)
    def _validate_cancelled_at(cls, value: Any):  # pylint: disable=no-self-argument
        return cls._validate_neo4j_datetime(value)

    @validator("progress")
    def _validate_progress(cls, value: float):
        # pylint: disable=no-self-argument
        if value is not None and not 0 <= value <= 1.0:
            msg = f"progress is expected to be in [0.0, 1.0], found {value}"
            raise ValueError(msg)
        return value

    @final
    @classmethod
    def from_neo4j(cls, record: "neo4j.Record", *, key: str = "task") -> Task:
        node = record[key]
        labels = node.labels
        node = dict(node)
        if len(labels) != 2:
            raise ValueError(f"Expected task to have exactly 2 labels found {labels}")
        state = [label for label in labels if label != NEO4J_TASK_NODE]
        if len(state) != 1:
            raise ValueError(f"Invalid task labels {labels}")
        state = state[0]
        if "completedAt" in node:
            node["completedAt"] = node["completedAt"].to_native()
        if "args" in node:
            node["args"] = json.loads(node["args"])
        if "group" in node:
            node.pop("group")
        node["state"] = state
        return cls(**node)

    @classmethod
    def postgres_row_factory(cls, cursor: "BaseCursor[Any, Any]") -> "RowMaker[Task]":

        def as_row(values: Sequence[Any]) -> Task:
            as_dict = {
                k.name: v
                for k, v in zip(cursor.description, values)
                if k.name != "group"
            }
            return cls(**as_dict)

        return as_row

    @final
    def resolve_event(self, event: TaskEvent) -> Optional[TaskUpdate]:
        if self.state in READY_STATES:
            return None
        updated = event.dict(exclude_unset=True, by_alias=False)
        for k in self._non_inherited_from_event:
            updated.pop(k, None)
        updated.pop(event.registry_key.default, None)
        base_update = TaskUpdate(**updated)
        # Update the state to make it consistent in case of race condition
        if isinstance(event, ProgressEvent):
            return self._progress_update(base_update)
        if isinstance(event, ErrorEvent):
            return self._error_update(base_update, event)
        if isinstance(event, CancelledEvent):
            return self._cancelled_update(base_update, event)
        if isinstance(event, ResultEvent):
            return self._result_update(base_update, event)
        raise TypeError(f"Unexpected event type {event.__class__}")

    def _result_update(self, base_update: TaskUpdate, event: ResultEvent) -> TaskUpdate:
        update = dict()
        update["progress"] = 1.0
        update["state"] = TaskState.DONE
        update["completed_at"] = event.created_at
        return safe_copy(base_update, update=update)

    def _cancelled_update(
        self, base_update: TaskUpdate, event: CancelledEvent
    ) -> TaskUpdate:
        update = dict()
        cancelled_state = TaskState.QUEUED if event.requeue else TaskState.CANCELLED
        update["state"] = cancelled_state
        if event.requeue:
            update["cancelled_at"] = None
            update["progress"] = 0.0
        else:
            update["cancelled_at"] = event.created_at
        updated = safe_copy(base_update, update=update)
        return updated

    def _error_update(self, base_update: TaskUpdate, event: ErrorEvent):
        update = dict()
        retries_left = min(self.retries_left, event.retries_left)
        can_retry = self.max_retries is not None and retries_left > 0
        update["state"] = TaskState.QUEUED if can_retry else TaskState.ERROR
        update["retries_left"] = retries_left
        updated = safe_copy(base_update, update=update)
        return updated

    def _progress_update(self, updated: TaskUpdate) -> TaskUpdate:
        state = TaskState.resolve_update_state(
            self, TaskUpdate(state=TaskState.RUNNING)
        )
        update = {"state": state}
        if state is TaskState.QUEUED:
            update["progress"] = 0.0
        return safe_copy(updated, update=update)

    def as_resolved(self, event: TaskEvent) -> Task:
        update = self.resolve_event(event)
        if update is None:
            return self
        return safe_copy(self, update=update.dict(exclude_unset=True))

    @final
    @classmethod
    def _schema(cls, by_alias: bool) -> Dict[str, Any]:
        global _TASK_SCHEMA
        if _TASK_SCHEMA is None:
            _TASK_SCHEMA = dict()
            _TASK_SCHEMA[True] = cls.schema(by_alias=True)
            _TASK_SCHEMA[False] = cls.schema(by_alias=False)
        return _TASK_SCHEMA[by_alias]


class StacktraceItem(LowerCamelCaseModel):
    name: str
    file: str
    lineno: int


@Message.register("TaskError")
class TaskError(Message, LowerCamelCaseModel):
    # Follow the "problem detail" spec: https://datatracker.ietf.org/doc/html/rfc9457,
    # the type is omitted for now since we gave no URI to resolve errors yet
    name: str
    message: str
    cause: Optional[str] = None
    stacktrace: List[StacktraceItem] = Field(default_factory=list)

    @classmethod
    def from_exception(cls, exception: BaseException) -> TaskError:
        name = exception.__class__.__name__
        message = str(exception)
        stacktrace = traceback.StackSummary.extract(
            traceback.walk_tb(exception.__traceback__)
        )
        stacktrace = [
            StacktraceItem(name=f.name, file=f.filename, lineno=f.lineno)
            for f in stacktrace
        ]
        cause = exception.__cause__
        if cause is not None:
            cause = str(cause)
        error = cls(name=name, message=message, cause=cause, stacktrace=stacktrace)
        return error


class TaskEvent(
    Message, NoEnumModel, LowerCamelCaseModel, Neo4jDatetimeMixin, FromTask, ABC
):
    task_id: str
    created_at: datetime
    retries_left: int = 3

    @validator("created_at", pre=True)
    def _validate_created_at(cls, value: Any):  # pylint: disable=no-self-argument
        return cls._validate_neo4j_datetime(value)


class WorkerEvent(TaskEvent, ABC):
    pass


class ManagerEvent(TaskEvent, ABC):
    pass


@Message.register("ProgressEvent")
class ProgressEvent(ManagerEvent):
    progress: float

    @validator("progress")
    def _validate_progress(cls, value: float):
        # pylint: disable=no-self-argument
        if not 0 <= value <= 1.0:
            msg = f"progress is expected to be in [0.0, 1.0], found {value}"
            raise ValueError(msg)
        return value

    @classmethod
    def from_task(cls, task: Task, **kwargs) -> ProgressEvent:
        created_at = datetime.now(timezone.utc)
        event = cls(
            task_id=task.id, progress=task.progress, created_at=created_at, **kwargs
        )
        return event


@Message.register("CancelEvent")
class CancelEvent(WorkerEvent):
    requeue: bool

    @classmethod
    def from_neo4j(
        cls, record: "neo4j.Record", *, event_key: str = "event", task_key: str = "task"
    ) -> CancelEvent:
        task = record.get(task_key)
        event = record.get(event_key)
        task_id = task[NEO4J_TASK_ID]
        requeue = event[NEO4J_TASK_CANCEL_EVENT_REQUEUE]
        created_at = event[NEO4J_TASK_CANCEL_EVENT_CANCELLED_AT]
        return cls(task_id=task_id, requeue=requeue, created_at=created_at)

    @classmethod
    def from_task(cls, task: Task, *, requeue: bool, **kwargs) -> CancelEvent:
        # pylint: disable=arguments-differ
        return cls(
            task_id=task.id, requeue=requeue, created_at=datetime.now(timezone.utc)
        )


@Message.register("CancelledEvent")
class CancelledEvent(ManagerEvent):
    requeue: bool

    @classmethod
    def from_task(cls, task: Task, *, requeue: bool, **kwargs) -> CancelledEvent:
        # pylint: disable=arguments-differ
        created_at = datetime.now(timezone.utc)
        event = cls(task_id=task.id, created_at=created_at, requeue=requeue)
        return event


@Message.register("ResultEvent")
class ResultEvent(ManagerEvent):
    result: object

    @classmethod
    def from_task(cls, task: Task, result: object, **kwargs) -> ResultEvent:
        # pylint: disable=arguments-differ
        return cls(
            task_id=task.id,
            result=result,
            created_at=datetime.now(timezone.utc),
            **kwargs,
        )

    @classmethod
    def from_neo4j(
        cls,
        record: "neo4j.Record",
        *,
        task_key: str = "task",
        result_key: str = "result",
    ) -> ResultEvent:
        result = record.get(result_key)
        if result is not None:
            result = json.loads(result[NEO4J_TASK_RESULT_RESULT])
        task_id = record[task_key][NEO4J_TASK_ID]
        completed_at = record[task_key][NEO4J_TASK_COMPLETED_AT]
        as_dict = {"result": result, "task_id": task_id, "created_at": completed_at}
        return ResultEvent(**as_dict)

    @classmethod
    def postgres_row_factory(
        cls, cursor: "BaseCursor[Any, Any]"
    ) -> "RowMaker[ResultEvent]":
        def as_row(values: Sequence[Any]) -> ResultEvent:
            # pylint: disable=c-extension-no-member
            import ujson

            as_dict = {k.name: v for k, v in zip(cursor.description, values)}
            as_dict[NEO4J_TASK_RESULT_RESULT] = ujson.loads(
                as_dict[NEO4J_TASK_RESULT_RESULT]
            )
            return cls(**as_dict)

        return as_row


@Message.register("ErrorEvent")
class ErrorEvent(ManagerEvent):
    error: TaskError

    @classmethod
    def from_task(
        cls,
        task: Task,
        error: TaskError,
        retries_left: int,
        created_at: datetime,
        **kwargs,
    ) -> ErrorEvent:
        # pylint: disable=arguments-differ
        return cls(
            task_id=task.id,
            error=error,
            retries_left=retries_left,
            created_at=created_at,
        )

    @classmethod
    def from_neo4j(
        cls,
        record: "neo4j.Record",
        *,
        task_key: str = "task",
        error_key: str = "error",
        rel_key: str = "rel",
    ) -> ErrorEvent:
        error = dict(record.value(error_key))
        if NEO4J_TASK_ERROR_STACKTRACE in error:
            stacktrace = [
                StacktraceItem(**json.loads(item))
                for item in error[NEO4J_TASK_ERROR_STACKTRACE]
            ]
            error[NEO4J_TASK_ERROR_STACKTRACE] = stacktrace
        error = TaskError(**error)
        rel = dict(record.value(rel_key))
        task_id = record[task_key][NEO4J_TASK_ID]
        retries_left = rel[NEO4J_TASK_ERROR_OCCURRED_TYPE_RETRIES_LEFT]
        created_at = rel[NEO4J_TASK_ERROR_OCCURRED_TYPE_OCCURRED_AT]
        return ErrorEvent(
            task_id=task_id,
            error=error,
            created_at=created_at,
            retries_left=retries_left,
        )

    @classmethod
    def postgres_row_factory(
        cls, cursor: "BaseCursor[Any, Any]"
    ) -> "RowMaker[ErrorEvent]":
        def as_row(values: Sequence[Any]) -> cls:
            # pylint: disable=c-extension-no-member
            import ujson

            as_dict = {k.name: v for k, v in zip(cursor.description, values)}
            stacktrace = ujson.loads(as_dict.pop(NEO4J_TASK_ERROR_STACKTRACE))
            message = as_dict.pop(NEO4J_TASK_ERROR_MESSAGE)
            name = as_dict.pop(NEO4J_TASK_ERROR_NAME)
            cause = as_dict.pop(TASK_ERROR_CAUSE)
            task_error = {
                "stacktrace": stacktrace,
                "name": name,
                "message": message,
                "cause": cause,
            }
            as_dict["error"] = task_error
            return cls(**as_dict)

        return as_row


class TaskUpdate(NoEnumModel, LowerCamelCaseModel, FromTask):
    state: Optional[TaskState] = None
    progress: Optional[float] = None
    retries_left: Optional[int] = None
    completed_at: Optional[datetime] = None
    cancelled_at: Optional[datetime] = None

    _from_task = ["state", "progress", "retries_left", "completed_at", "cancelled_at"]

    @classmethod
    def from_task(cls, task: Task, **kwargs) -> TaskUpdate:
        from_task = {attr: getattr(task, attr) for attr in cls._from_task}
        from_task = {k: v for k, v in from_task.items() if v is not None}
        return cls(**from_task)

    @classmethod
    @lru_cache(maxsize=1)
    def done(cls, completed_at: Optional[datetime] = None) -> TaskUpdate:
        return cls(progress=1.0, completed_at=completed_at, state=TaskState.DONE)


def _id_title(title: str) -> str:
    id_title = []
    for i, letter in enumerate(title):
        if i and letter.isupper():
            id_title.append("-")
        id_title.append(letter.lower())
    return "".join(id_title)


Registrable.update_forward_refs()



from __future__ import annotations

import re
from contextlib import AbstractAsyncContextManager, AsyncExitStack, asynccontextmanager
from copy import deepcopy
from enum import Enum, unique
from functools import cached_property, lru_cache
from typing import Any, AsyncContextManager, Dict, List, Optional, Tuple, Type, cast
from urllib import parse

from aio_pika import (
    DeliveryMode,
    Message as AioPikaMessage,
    RobustChannel as RobustChannel_,
    RobustConnection as RobustConnection_,
)
from aio_pika.abc import (
    AbstractExchange,
    AbstractQueueIterator,
    AbstractRobustChannel,
    AbstractRobustConnection,
    ExchangeType,
)
from aio_pika.exceptions import ChannelPreconditionFailed
from aiohttp import (
    BasicAuth,
    ClientResponse,
    ClientResponseError,
    ClientSession,
)
from aiohttp.client import _RequestOptions
from aiohttp.typedefs import StrOrURL
from aiormq.abc import ConfirmationFrameType
from pamqp.commands import Basic
from typing_extensions import Unpack

from icij_common.pydantic_utils import ICIJModel, NoEnumModel
from icij_worker import Message
from icij_worker.app import TaskGroup
from icij_worker.constants import (
    AMQP_MANAGER_EVENTS_DL_QUEUE,
    AMQP_MANAGER_EVENTS_DL_ROUTING_KEY,
    AMQP_MANAGER_EVENTS_DL_X,
    AMQP_MANAGER_EVENTS_QUEUE,
    AMQP_MANAGER_EVENTS_ROUTING_KEY,
    AMQP_MANAGER_EVENTS_X,
    AMQP_TASKS_DL_QUEUE,
    AMQP_TASKS_DL_ROUTING_KEY,
    AMQP_TASKS_DL_X,
    AMQP_TASKS_QUEUE,
    AMQP_TASKS_ROUTING_KEY,
    AMQP_TASKS_X,
    AMQP_TASK_QUEUE_PRIORITY,
    AMQP_WORKER_EVENTS_QUEUE,
    AMQP_WORKER_EVENTS_ROUTING_KEY,
    AMQP_WORKER_EVENTS_X,
)
from icij_worker.exceptions import WorkerTimeoutError
from icij_worker.routing_strategy import Exchange, Routing, RoutingStrategy
from icij_worker.task_storage.postgres.postgres import logger

_DELIVERY_ACK_TIMEOUT_RE = re.compile(
    r"delivery acknowledgement on channel .+ timed out", re.MULTILINE
)


@unique
class ApplyTo(str, Enum):
    EXCHANGES = "exchanges"
    QUEUES = "queues"
    CLASSIC_QUEUES = "classic_queues"
    QUORUM_QUEUES = "quorum_queues"
    STREAMS = "streams"
    ALL = "all"


class AMQPPolicy(NoEnumModel):
    name: str
    pattern: str
    definition: Dict[str, Any]
    apply_to: Optional[ApplyTo] = None
    priority: Optional[int] = None


_DELIVERY_ACK_TIMEOUT_RE = re.compile(
    r"delivery acknowledgement on channel .+ timed out", re.MULTILINE
)


class AMQPConfigMixin(ICIJModel):
    connection_timeout_s: float = 5.0
    reconnection_wait_s: float = 5.0
    rabbitmq_host: str = "127.0.0.1"
    rabbitmq_password: str = "guest"
    rabbitmq_port: Optional[int] = 5672
    rabbitmq_management_port: Optional[int] = 15672
    rabbitmq_user: Optional[str] = "guest"
    rabbitmq_vhost: Optional[str] = "%2F"

    @cached_property
    def broker_url(self) -> str:
        amqp_userinfo = self.rabbitmq_user
        amqp_userinfo += f":{self.rabbitmq_password}"
        if amqp_userinfo:
            amqp_userinfo += "@"
        amqp_authority = (
            f"{amqp_userinfo or ''}{self.rabbitmq_host}"
            f"{f':{self.rabbitmq_port}' or ''}"
        )
        amqp_uri = f"amqp://{amqp_authority}"
        if self.rabbitmq_vhost is not None:
            amqp_uri += f"/{self.rabbitmq_vhost}"
        return amqp_uri

    @cached_property
    def management_url(self) -> str:
        management_url = f"http://{self.rabbitmq_host}:{self.rabbitmq_management_port}"
        return management_url

    @cached_property
    def basic_auth(self) -> BasicAuth:
        return BasicAuth(self.rabbitmq_user, self.rabbitmq_password)

    def to_management_client(self) -> AMQPManagementClient:
        client = AMQPManagementClient(
            self.management_url,
            rabbitmq_vhost=self.rabbitmq_vhost,
            rabbitmq_auth=self.basic_auth,
        )
        return client


class AMQPMixin:
    _app_id: str
    _channel_: AbstractRobustChannel
    _routing_strategy: RoutingStrategy
    _task_x: AbstractExchange
    max_task_queue_size: Optional[int]
    _always_include = {"createdAt", "retriesLeft"}

    def __init__(
        self,
        broker_url: str,
        *,
        connection_timeout_s: float = 1.0,
        reconnection_wait_s: float = 5.0,
        inactive_after_s: float = None,
    ):
        self._broker_url = broker_url
        self._reconnection_wait_s = reconnection_wait_s
        self._connection_timeout_s = connection_timeout_s
        self._inactive_after_s = inactive_after_s
        self._connection_: Optional[AbstractRobustConnection] = None
        self._exit_stack = AsyncExitStack()

    async def _publish_message(
        self,
        message: Message,
        *,
        exchange: AbstractExchange,
        delivery_mode: DeliveryMode = DeliveryMode.PERSISTENT,
        routing_key: Optional[str],
        mandatory: bool,
    ) -> Optional[ConfirmationFrameType]:
        message = message.json(
            exclude_unset=True, by_alias=True, exclude_none=True
        ).encode()
        message = AioPikaMessage(
            message, delivery_mode=delivery_mode, app_id=self._app_id
        )
        confirmation = await exchange.publish(message, routing_key, mandatory=mandatory)
        if not isinstance(confirmation, Basic.Ack):
            msg = f"Failed to deliver {message.body}, received {confirmation}"
            raise RuntimeError(msg)
        return confirmation

    @property
    def _connection(self) -> AbstractRobustConnection:
        if self._connection_ is None:
            msg = (
                f"{self} has no connection, please call"
                f" {self.__class__.__aenter__.__name__}"
            )
            raise ValueError(msg)
        return self._connection_

    @property
    def _channel(self) -> AbstractRobustChannel:
        if self._channel_ is None:
            msg = (
                f"{self} has no channel, please call"
                f" {self.__class__.__aenter__.__name__}"
            )
            raise ValueError(msg)
        return self._channel_

    @property
    def channel(self) -> AbstractRobustChannel:
        return self._channel

    @property
    def connection(self) -> AbstractRobustChannel:
        return self._connection

    @classmethod
    @lru_cache(maxsize=1)
    def default_task_routing(cls) -> Routing:
        return Routing(
            exchange=Exchange(name=AMQP_TASKS_X, type=ExchangeType.DIRECT),
            routing_key=AMQP_TASKS_ROUTING_KEY,
            queue_name=AMQP_TASKS_QUEUE,
            queue_args={"x-queue-type": "quorum"},
            dead_letter_routing=Routing(
                exchange=Exchange(name=AMQP_TASKS_DL_X, type=ExchangeType.DIRECT),
                routing_key=AMQP_TASKS_DL_ROUTING_KEY,
                queue_name=AMQP_TASKS_DL_QUEUE,
            ),
        )

    @classmethod
    @lru_cache(maxsize=1)
    def manager_evt_routing(cls) -> Routing:
        return Routing(
            exchange=Exchange(name=AMQP_MANAGER_EVENTS_X, type=ExchangeType.DIRECT),
            routing_key=AMQP_MANAGER_EVENTS_ROUTING_KEY,
            queue_name=AMQP_MANAGER_EVENTS_QUEUE,
            dead_letter_routing=Routing(
                exchange=Exchange(
                    name=AMQP_MANAGER_EVENTS_DL_X, type=ExchangeType.DIRECT
                ),
                routing_key=AMQP_MANAGER_EVENTS_DL_ROUTING_KEY,
                queue_name=AMQP_MANAGER_EVENTS_DL_QUEUE,
            ),
        )

    @classmethod
    @lru_cache(maxsize=1)
    def worker_evt_routing(cls) -> Routing:
        return Routing(
            exchange=Exchange(name=AMQP_WORKER_EVENTS_X, type=ExchangeType.FANOUT),
            routing_key=AMQP_WORKER_EVENTS_ROUTING_KEY,
            queue_name=AMQP_WORKER_EVENTS_QUEUE,
        )

    async def _get_queue_iterator(
        self,
        routing: Routing,
        *,
        declare_exchanges: bool,
        declare_queues: bool = True,
        durable_queues: bool = True,
    ) -> Tuple[AbstractQueueIterator, AbstractExchange, Optional[AbstractExchange]]:
        await self._exit_stack.enter_async_context(
            cast(AbstractAsyncContextManager, self._channel)
        )
        dlq_ex = None
        await self._create_routing(
            routing,
            declare_exchanges=declare_exchanges,
            declare_queues=declare_queues,
            durable_queues=durable_queues,
        )
        ex = await self._channel.get_exchange(routing.exchange.name, ensure=True)
        queue = await self._channel.get_queue(routing.queue_name, ensure=True)
        kwargs = dict()
        if self._inactive_after_s is not None:
            kwargs["timeout"] = self._inactive_after_s
        return queue.iterator(**kwargs), ex, dlq_ex

    async def _create_routing(
        self,
        routing: Routing,
        *,
        declare_exchanges: bool = True,
        declare_queues: bool = True,
        durable_queues: bool = True,
    ):
        if declare_exchanges:
            x = await self._channel.declare_exchange(
                routing.exchange.name, type=routing.exchange.type, durable=True
            )
        else:
            x = await self._channel.get_exchange(routing.exchange.name, ensure=True)
        queue_args = None
        if routing.queue_args is not None:
            queue_args = deepcopy(routing.queue_args)
        if routing.dead_letter_routing:
            await self._create_routing(
                routing.dead_letter_routing,
                declare_exchanges=declare_exchanges,
                declare_queues=declare_queues,
                durable_queues=durable_queues,
            )
            if queue_args is None:
                queue_args = dict()
            dlx_name = routing.dead_letter_routing.exchange.name
            dl_routing_key = routing.dead_letter_routing.routing_key
            # TODO: this could be passed through policy
            update = {
                "x-dead-letter-exchange": dlx_name,
                "x-dead-letter-routing-key": dl_routing_key,
            }
            queue_args.update(update)
        if declare_queues:
            queue = await self._channel.declare_queue(
                routing.queue_name, durable=durable_queues, arguments=queue_args
            )
        else:
            queue = await self._channel.get_queue(routing.queue_name, ensure=True)
        await queue.bind(x, routing_key=routing.routing_key)


class AMQPManagementClient(AsyncContextManager):
    def __init__(
        self,
        rabbitmq_management_url: str,
        *,
        rabbitmq_vhost: str,
        rabbitmq_auth: BasicAuth,
    ):
        self._management_url = rabbitmq_management_url
        self._vhost = rabbitmq_vhost
        self._auth = rabbitmq_auth
        self._session: Optional[ClientSession]

    async def __aenter__(self):
        self._session = ClientSession(self._management_url, auth=self._auth)
        await self._session.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self._session.__aexit__(exc_type, exc_value, traceback)

    @asynccontextmanager
    async def _put(self, url: StrOrURL, *, data: Any = None, **kwargs: Any):
        async with self._session.put(url, data=data, **kwargs) as res:
            _raise_for_status(res)
            yield res

    @asynccontextmanager
    async def _get(self, url: StrOrURL, *, allow_redirects: bool = True, **kwargs: Any):
        async with self._session.get(
            url, allow_redirects=allow_redirects, **kwargs
        ) as res:
            _raise_for_status(res)
            yield res

    @asynccontextmanager
    async def _delete(self, url: StrOrURL, **kwargs: Unpack[_RequestOptions]):
        async with self._session.delete(url, **kwargs) as res:
            _raise_for_status(res)
            yield res

    async def set_policy(self, policy: AMQPPolicy):
        url = f"/api/policies/{self._vhost}/{parse.quote(policy.name)}"
        data = {"pattern": policy.pattern, "definition": policy.definition}
        if policy.apply_to is not None:
            data["apply-to"] = policy.apply_to.value
        if policy.priority is not None:
            data["priority"] = policy.priority
        async with self._put(url, json=data):
            pass

    async def list_policies(self) -> List[Dict]:
        url = f"/api/policies/{self._vhost}"
        async with self._get(url) as res:
            return await res.json()

    async def clear_policies(self):
        policies = await self.list_policies()
        for p in policies:
            url = f"/api/policies/{self._vhost}/{p['name']}"
            async with self._delete(url):
                pass


def amqp_task_group_policy(
    task_routing: Routing,
    group: Optional[TaskGroup],
    app_max_task_queue_size: Optional[int],
) -> AMQPPolicy:
    pattern = rf"^{re.escape(task_routing.queue_name)}$"
    name = f"task-group-policy-{task_routing.queue_name}"
    definition = {"overflow": "reject-publish", "delivery-limit": 10}
    max_task_queue_size = app_max_task_queue_size
    if group is not None and group.max_task_queue_size is not None:
        max_task_queue_size = group.max_task_queue_size
    if max_task_queue_size is not None:
        definition["max-length"] = max_task_queue_size
    return AMQPPolicy(
        name=name,
        pattern=pattern,
        definition=definition,
        apply_to=ApplyTo.QUEUES,
        priority=AMQP_TASK_QUEUE_PRIORITY,
    )


def _raise_for_status(res: ClientResponse):
    try:
        res.raise_for_status()
    except ClientResponseError as e:
        msg = "request to %s, failed with reason %s"
        logger.exception(msg, res.request_info, res.reason)
        raise e


class RobustChannel(RobustChannel_):
    async def __close_callback(self, _: Any, exc: BaseException) -> None:
        # pylint: disable=unused-private-member
        timeout_exc = parse_consumer_timeout(exc)
        if timeout_exc is not None:
            logger.error("channel closing due to consumer timeout: %s", exc)
            raise timeout_exc from exc


class RobustConnection(RobustConnection_):
    CHANNEL_CLASS: Type[RobustChannel] = RobustChannel

    # Defined async context manager attributes to be able to enter and exit this
    # in ExitStack
    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()


def parse_consumer_timeout(exc: BaseException) -> Optional[WorkerTimeoutError]:
    if not isinstance(exc, ChannelPreconditionFailed):
        return None
    if not exc.args:
        return None
    msg = exc.args[0]
    if _DELIVERY_ACK_TIMEOUT_RE.search(msg):
        return WorkerTimeoutError(msg)
    return None
